# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library of losses for disentanglement learning.

Implementation of VAE based models for unsupervised learning of disentangled
representations.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import math
from disentanglement_lib.methods.shared import architectures  # pylint: disable=unused-import
from disentanglement_lib.methods.shared import losses  # pylint: disable=unused-import
from disentanglement_lib.methods.shared import optimizers  # pylint: disable=unused-import
from disentanglement_lib.methods.unsupervised import gaussian_encoder_model
from six.moves import range
from six.moves import zip
import tensorflow.compat.v1 as tf
import gin.tf
from tensorflow.contrib import tpu as contrib_tpu


class BaseVAE(gaussian_encoder_model.GaussianEncoderModel):
  """Abstract base class of a basic Gaussian encoder model."""

  def model_fn(self, features, labels, mode, params):
    """TPUEstimator compatible model function."""
    del labels
    is_training = (mode == tf.estimator.ModeKeys.TRAIN)
    data_shape = features.get_shape().as_list()[1:]
    z_mean, z_logvar = self.gaussian_encoder(features, is_training=is_training)
    z_sampled = self.sample_from_latent_distribution(z_mean, z_logvar)
    reconstructions = self.decode(z_sampled, data_shape, is_training)
    per_sample_loss = losses.make_reconstruction_loss(features, reconstructions)
    reconstruction_loss = tf.reduce_mean(per_sample_loss)
    kl_loss = compute_gaussian_kl(z_mean, z_logvar)
    regularizer = self.regularizer(kl_loss, z_mean, z_logvar, z_sampled)
    loss = tf.add(reconstruction_loss, regularizer, name="loss")
    elbo = tf.add(reconstruction_loss, kl_loss, name="elbo")
    if mode == tf.estimator.ModeKeys.TRAIN:
      optimizer = optimizers.make_vae_optimizer()
      update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
      train_op = optimizer.minimize(
          loss=loss, global_step=tf.train.get_global_step())
      train_op = tf.group([train_op, update_ops])
      tf.summary.scalar("reconstruction_loss", reconstruction_loss)
      tf.summary.scalar("elbo", -elbo)

      logging_hook = tf.train.LoggingTensorHook({
          "loss": loss,
          "reconstruction_loss": reconstruction_loss,
          "elbo": -elbo
      },
                                                every_n_iter=100)
      return contrib_tpu.TPUEstimatorSpec(
          mode=mode,
          loss=loss,
          train_op=train_op,
          training_hooks=[logging_hook])
    elif mode == tf.estimator.ModeKeys.EVAL:
      return contrib_tpu.TPUEstimatorSpec(
          mode=mode,
          loss=loss,
          eval_metrics=(make_metric_fn("reconstruction_loss", "elbo",
                                       "regularizer", "kl_loss"),
                        [reconstruction_loss, -elbo, regularizer, kl_loss]))
    else:
      raise NotImplementedError("Eval mode not supported.")

  def gaussian_encoder(self, input_tensor, is_training):
    """Applies the Gaussian encoder to images.

    Args:
      input_tensor: Tensor with the observations to be encoded.
      is_training: Boolean indicating whether in training mode.

    Returns:
      Tuple of tensors with the mean and log variance of the Gaussian encoder.
    """
    return architectures.make_gaussian_encoder(
        input_tensor, is_training=is_training)

  def decode(self, latent_tensor, observation_shape, is_training):
    """Decodes the latent_tensor to an observation."""
    return architectures.make_decoder(
        latent_tensor, observation_shape, is_training=is_training)


def shuffle_codes(z):
  """Shuffles latent variables across the batch.

  Args:
    z: [batch_size, num_latent] representation.

  Returns:
    shuffled: [batch_size, num_latent] shuffled representation across the batch.
  """
  z_shuffle = []
  for i in range(z.get_shape()[1]):
    z_shuffle.append(tf.random_shuffle(z[:, i]))
  shuffled = tf.stack(z_shuffle, 1, name="latent_shuffled")
  return shuffled


def compute_gaussian_kl(z_mean, z_logvar):
  """Compute KL divergence between input Gaussian and Standard Normal."""
  return tf.reduce_mean(
      0.5 * tf.reduce_sum(
          tf.square(z_mean) + tf.exp(z_logvar) - z_logvar - 1, [1]),
      name="kl_loss")


def make_metric_fn(*names):
  """Utility function to report tf.metrics in model functions."""

  def metric_fn(*args):
    return {name: tf.metrics.mean(vec) for name, vec in zip(names, args)}

  return metric_fn


@gin.configurable("vae")
class BetaVAE(BaseVAE):
  """BetaVAE model."""

  def __init__(self, beta=gin.REQUIRED):
    """Creates a beta-VAE model.

    Implementing Eq. 4 of "beta-VAE: Learning Basic Visual Concepts with a
    Constrained Variational Framework"
    (https://openreview.net/forum?id=Sy2fzU9gl).

    Args:
      beta: Hyperparameter for the regularizer.

    Returns:
      model_fn: Model function for TPUEstimator.
    """
    self.beta = beta

  def regularizer(self, kl_loss, z_mean, z_logvar, z_sampled):
    del z_mean, z_logvar, z_sampled
    return self.beta * kl_loss


def anneal(c_max, step, iteration_threshold):
  """Anneal function for anneal_vae (https://arxiv.org/abs/1804.03599).

  Args:
    c_max: Maximum capacity.
    step: Current step.
    iteration_threshold: How many iterations to reach c_max.

  Returns:
    Capacity annealed linearly until c_max.
  """
  return tf.math.minimum(c_max * 1.,
                         c_max * 1. * tf.to_float(step) / iteration_threshold)


@gin.configurable("annealed_vae")
class AnnealedVAE(BaseVAE):
  """AnnealedVAE model."""

  def __init__(self,
               gamma=gin.REQUIRED,
               c_max=gin.REQUIRED,
               iteration_threshold=gin.REQUIRED):
    """Creates an AnnealedVAE model.

    Implementing Eq. 8 of "Understanding disentangling in beta-VAE"
    (https://arxiv.org/abs/1804.03599).

    Args:
      gamma: Hyperparameter for the regularizer.
      c_max: Maximum capacity of the bottleneck.
      iteration_threshold: How many iterations to reach c_max.
    """
    self.gamma = gamma
    self.c_max = c_max
    self.iteration_threshold = iteration_threshold

  def regularizer(self, kl_loss, z_mean, z_logvar, z_sampled):
    del z_mean, z_logvar, z_sampled
    c = anneal(self.c_max, tf.train.get_global_step(), self.iteration_threshold)
    return self.gamma * tf.math.abs(kl_loss - c)


@gin.configurable("factor_vae")
class FactorVAE(BaseVAE):
  """FactorVAE model."""

  def __init__(self, gamma=gin.REQUIRED):
    """Creates a FactorVAE model.

    Implementing Eq. 2 of "Disentangling by Factorizing"
    (https://arxiv.org/pdf/1802.05983).

    Args:
      gamma: Hyperparameter for the regularizer.
    """
    self.gamma = gamma

  def model_fn(self, features, labels, mode, params):
    """TPUEstimator compatible model function."""
    del labels
    is_training = (mode == tf.estimator.ModeKeys.TRAIN)
    data_shape = features.get_shape().as_list()[1:]
    z_mean, z_logvar = self.gaussian_encoder(features, is_training=is_training)
    z_sampled = self.sample_from_latent_distribution(z_mean, z_logvar)
    z_shuffle = shuffle_codes(z_sampled)
    with tf.variable_scope(tf.get_variable_scope(), reuse=tf.AUTO_REUSE):
      logits_z, probs_z = architectures.make_discriminator(
          z_sampled, is_training=is_training)
      _, probs_z_shuffle = architectures.make_discriminator(
          z_shuffle, is_training=is_training)
    reconstructions = self.decode(z_sampled, data_shape, is_training)
    per_sample_loss = losses.make_reconstruction_loss(
        features, reconstructions)
    reconstruction_loss = tf.reduce_mean(per_sample_loss)
    kl_loss = compute_gaussian_kl(z_mean, z_logvar)
    standard_vae_loss = tf.add(reconstruction_loss, kl_loss, name="VAE_loss")
    # tc = E[log(p_real)-log(p_fake)] = E[logit_real - logit_fake]
    tc_loss_per_sample = logits_z[:, 0] - logits_z[:, 1]
    tc_loss = tf.reduce_mean(tc_loss_per_sample, axis=0)
    regularizer = kl_loss + self.gamma * tc_loss
    factor_vae_loss = tf.add(
        standard_vae_loss, self.gamma * tc_loss, name="factor_VAE_loss")
    discr_loss = tf.add(
        0.5 * tf.reduce_mean(tf.log(probs_z[:, 0])),
        0.5 * tf.reduce_mean(tf.log(probs_z_shuffle[:, 1])),
        name="discriminator_loss")
    if mode == tf.estimator.ModeKeys.TRAIN:
      optimizer_vae = optimizers.make_vae_optimizer()
      optimizer_discriminator = optimizers.make_discriminator_optimizer()
      all_variables = tf.trainable_variables()
      encoder_vars = [var for var in all_variables if "encoder" in var.name]
      decoder_vars = [var for var in all_variables if "decoder" in var.name]
      discriminator_vars = [var for var in all_variables \
                            if "discriminator" in var.name]
      update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
      train_op_vae = optimizer_vae.minimize(
          loss=factor_vae_loss,
          global_step=tf.train.get_global_step(),
          var_list=encoder_vars + decoder_vars)
      train_op_discr = optimizer_discriminator.minimize(
          loss=-discr_loss,
          global_step=tf.train.get_global_step(),
          var_list=discriminator_vars)
      train_op = tf.group(train_op_vae, train_op_discr, update_ops)
      tf.summary.scalar("reconstruction_loss", reconstruction_loss)
      logging_hook = tf.train.LoggingTensorHook({
          "loss": factor_vae_loss,
          "reconstruction_loss": reconstruction_loss
      },
                                                every_n_iter=50)
      return contrib_tpu.TPUEstimatorSpec(
          mode=mode,
          loss=factor_vae_loss,
          train_op=train_op,
          training_hooks=[logging_hook])
    elif mode == tf.estimator.ModeKeys.EVAL:
      return contrib_tpu.TPUEstimatorSpec(
          mode=mode,
          loss=factor_vae_loss,
          eval_metrics=(make_metric_fn("reconstruction_loss", "regularizer",
                                       "kl_loss"),
                        [reconstruction_loss, regularizer, kl_loss]))
    else:
      raise NotImplementedError("Eval mode not supported.")


def compute_covariance_z_mean(z_mean):
  """Computes the covariance of z_mean.

  Uses cov(z_mean) = E[z_mean*z_mean^T] - E[z_mean]E[z_mean]^T.

  Args:
    z_mean: Encoder mean, tensor of size [batch_size, num_latent].

  Returns:
    cov_z_mean: Covariance of encoder mean, tensor of size [num_latent,
      num_latent].
  """
  expectation_z_mean_z_mean_t = tf.reduce_mean(
      tf.expand_dims(z_mean, 2) * tf.expand_dims(z_mean, 1), axis=0)
  expectation_z_mean = tf.reduce_mean(z_mean, axis=0)
  cov_z_mean = tf.subtract(
      expectation_z_mean_z_mean_t,
      tf.expand_dims(expectation_z_mean, 1) * tf.expand_dims(
          expectation_z_mean, 0))
  return cov_z_mean


def regularize_diag_off_diag_dip(covariance_matrix, lambda_od, lambda_d):
  """Compute on and off diagonal regularizers for DIP-VAE models.

  Penalize deviations of covariance_matrix from the identity matrix. Uses
  different weights for the deviations of the diagonal and off diagonal entries.

  Args:
    covariance_matrix: Tensor of size [num_latent, num_latent] to regularize.
    lambda_od: Weight of penalty for off diagonal elements.
    lambda_d: Weight of penalty for diagonal elements.

  Returns:
    dip_regularizer: Regularized deviation from diagonal of covariance_matrix.
  """
  covariance_matrix_diagonal = tf.diag_part(covariance_matrix)
  covariance_matrix_off_diagonal = covariance_matrix - tf.diag(
      covariance_matrix_diagonal)
  dip_regularizer = tf.add(
      lambda_od * tf.reduce_sum(covariance_matrix_off_diagonal**2),
      lambda_d * tf.reduce_sum((covariance_matrix_diagonal - 1)**2))
  return dip_regularizer


@gin.configurable("dip_vae")
class DIPVAE(BaseVAE):
  """DIPVAE model."""

  def __init__(self,
               lambda_od=gin.REQUIRED,
               lambda_d_factor=gin.REQUIRED,
               dip_type="i"):
    """Creates a DIP-VAE model.

    Based on Equation 6 and 7 of "Variational Inference of Disentangled Latent
    Concepts from Unlabeled Observations"
    (https://openreview.net/pdf?id=H1kG7GZAW).

    Args:
      lambda_od: Hyperparameter for off diagonal values of covariance matrix.
      lambda_d_factor: Hyperparameter for diagonal values of covariance matrix
        lambda_d = lambda_d_factor*lambda_od.
      dip_type: "i" or "ii".
    """
    self.lambda_od = lambda_od
    self.lambda_d_factor = lambda_d_factor
    self.dip_type = dip_type

  def regularizer(self, kl_loss, z_mean, z_logvar, z_sampled):
    cov_z_mean = compute_covariance_z_mean(z_mean)
    lambda_d = self.lambda_d_factor * self.lambda_od
    if self.dip_type == "i":  # Eq 6 page 4
      # mu = z_mean is [batch_size, num_latent]
      # Compute cov_p(x) [mu(x)] = E[mu*mu^T] - E[mu]E[mu]^T]
      cov_dip_regularizer = regularize_diag_off_diag_dip(
          cov_z_mean, self.lambda_od, lambda_d)
    elif self.dip_type == "ii":
      cov_enc = tf.matrix_diag(tf.exp(z_logvar))
      expectation_cov_enc = tf.reduce_mean(cov_enc, axis=0)
      cov_z = expectation_cov_enc + cov_z_mean
      cov_dip_regularizer = regularize_diag_off_diag_dip(
          cov_z, self.lambda_od, lambda_d)
    else:
      raise NotImplementedError("DIP variant not supported.")
    return kl_loss + cov_dip_regularizer


def gaussian_log_density(samples, mean, log_var):
  pi = tf.constant(math.pi)
  normalization = tf.log(2. * pi)
  inv_sigma = tf.exp(-log_var)
  tmp = (samples - mean)
  return -0.5 * (tmp * tmp * inv_sigma + log_var + normalization)


def total_correlation(z, z_mean, z_logvar):
  """Estimate of total correlation on a batch.

  We need to compute the expectation over a batch of: E_j [log(q(z(x_j))) -
  log(prod_l q(z(x_j)_l))]. We ignore the constants as they do not matter
  for the minimization. The constant should be equal to (num_latents - 1) *
  log(batch_size * dataset_size)

  Args:
    z: [batch_size, num_latents]-tensor with sampled representation.
    z_mean: [batch_size, num_latents]-tensor with mean of the encoder.
    z_logvar: [batch_size, num_latents]-tensor with log variance of the encoder.

  Returns:
    Total correlation estimated on a batch.
  """
  # Compute log(q(z(x_j)|x_i)) for every sample in the batch, which is a
  # tensor of size [batch_size, batch_size, num_latents]. In the following
  # comments, [batch_size, batch_size, num_latents] are indexed by [j, i, l].
  log_qz_prob = gaussian_log_density(
      tf.expand_dims(z, 1), tf.expand_dims(z_mean, 0),
      tf.expand_dims(z_logvar, 0))
  # Compute log prod_l p(z(x_j)_l) = sum_l(log(sum_i(q(z(z_j)_l|x_i)))
  # + constant) for each sample in the batch, which is a vector of size
  # [batch_size,].
  log_qz_product = tf.reduce_sum(
      tf.reduce_logsumexp(log_qz_prob, axis=1, keepdims=False),
      axis=1,
      keepdims=False)
  # Compute log(q(z(x_j))) as log(sum_i(q(z(x_j)|x_i))) + constant =
  # log(sum_i(prod_l q(z(x_j)_l|x_i))) + constant.
  log_qz = tf.reduce_logsumexp(
      tf.reduce_sum(log_qz_prob, axis=2, keepdims=False),
      axis=1,
      keepdims=False)
  return tf.reduce_mean(log_qz - log_qz_product)


@gin.configurable("beta_tc_vae")
class BetaTCVAE(BaseVAE):
  """BetaTCVAE model."""

  def __init__(self, beta=gin.REQUIRED):
    """Creates a beta-TC-VAE model.

    Based on Equation 4 with alpha = gamma = 1 of "Isolating Sources of
    Disentanglement in Variational Autoencoders"
    (https://arxiv.org/pdf/1802.04942).
    If alpha = gamma = 1, Eq. 4 can be written as ELBO + (1 - beta) * TC.

    Args:
      beta: Hyperparameter total correlation.
    """
    self.beta = beta

  def regularizer(self, kl_loss, z_mean, z_logvar, z_sampled):
    tc = (self.beta - 1.) * total_correlation(z_sampled, z_mean, z_logvar)
    return tc + kl_loss



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for fairness.py."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Dependency imports

from absl.testing import absltest
from disentanglement_lib.data.ground_truth import dummy_data
from disentanglement_lib.evaluation.metrics import fairness
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
import gin.tf


class FairnessTest(absltest.TestCase):

  def test_metric(self):
    gin.bind_parameter("predictor.predictor_fn",
                       utils.gradient_boosting_classifier)
    ground_truth_data = dummy_data.DummyData()
    def representation_function(x):
      return np.array(x, dtype=np.float64)[:, :, 0, 0]
    random_state = np.random.RandomState(0)
    _ = fairness.compute_fairness(ground_truth_data, representation_function,
                                  random_state, None, 1000, 1000)

  def test_inter_group_fairness(self):
    counts = np.array([[0, 100], [100, 100]])
    mean_fairness, max_fairness = fairness.inter_group_fairness(counts)
    # The overall distribution is 1/3 to 2/3.
    # The first sensitive class has a distribution of 0 to 1.
    # The second sensitive class has a distribution of 1/2 to 1/2.
    # The total variation distances are hence 1/3 and 1/6 respectively.
    # The mean fairness is hence 1/3*1/3 + 2/3*1/6 = 2/9.
    self.assertAlmostEqual(mean_fairness, 2. / 9.)
    self.assertAlmostEqual(max_fairness, 1 / 3.)

  def test_compute_scores_dict(self):
    scores = np.array([[0., 2., 4.], [4., 0., 8.], [2., 10., 0.]])
    results = fairness.compute_scores_dict(scores, "test")
    shouldbe = {
        # Single scores.
        "test:pred0:sens1": 2.,
        "test:pred0:sens2": 4.,
        "test:pred1:sens0": 4.,
        "test:pred1:sens2": 8.,
        "test:pred2:sens0": 2.,
        "test:pred2:sens1": 10.,
        # Column scores.
        "test:pred0:mean_sens": 3.,
        "test:pred0:max_sens": 4.,
        "test:pred1:mean_sens": 6.,
        "test:pred1:max_sens": 8.,
        "test:pred2:mean_sens": 6.,
        "test:pred2:max_sens": 10.,
        # Column scores.
        "test:sens0:mean_pred": 3.,
        "test:sens0:max_pred": 4.,
        "test:sens1:mean_pred": 6.,
        "test:sens1:max_pred": 10.,
        "test:sens2:mean_pred": 6.,
        "test:sens2:max_pred": 8.,
        # All scores.
        "test:mean_sens:mean_pred": 5.,
        "test:mean_sens:max_pred": 22. / 3.,
        "test:max_sens:mean_pred": 6.,
        "test:max_sens:max_pred": 10.,
        "test:mean_pred:mean_sens": 5.,
        "test:mean_pred:max_sens": 22. / 3.,
        "test:max_pred:mean_sens": 6.,
        "test:max_pred:max_sens": 10.,
    }
    self.assertDictEqual(shouldbe, results)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of Disentanglement, Completeness and Informativeness.

Based on "A Framework for the Quantitative Evaluation of Disentangled
Representations" (https://openreview.net/forum?id=By-7dz-AZ).
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl import logging
from disentanglement_lib.evaluation.metrics import utils
import numpy as np
import scipy
from six.moves import range
from sklearn import ensemble
import gin.tf


@gin.configurable(
    "dci",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_dci(ground_truth_data, representation_function, random_state,
                artifact_dir=None,
                num_train=gin.REQUIRED,
                num_test=gin.REQUIRED,
                batch_size=16):
  """Computes the DCI scores according to Sec 2.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    num_train: Number of points used for training.
    num_test: Number of points used for testing.
    batch_size: Batch size for sampling.

  Returns:
    Dictionary with average disentanglement score, completeness and
      informativeness (train and test).
  """
  del artifact_dir
  logging.info("Generating training set.")
  # mus_train are of shape [num_codes, num_train], while ys_train are of shape
  # [num_factors, num_train].
  mus_train, ys_train = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_train,
      random_state, batch_size)
  assert mus_train.shape[1] == num_train
  assert ys_train.shape[1] == num_train
  mus_test, ys_test = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_test,
      random_state, batch_size)
  scores = _compute_dci(mus_train, ys_train, mus_test, ys_test)
  return scores


def _compute_dci(mus_train, ys_train, mus_test, ys_test):
  """Computes score based on both training and testing codes and factors."""
  scores = {}
  importance_matrix, train_err, test_err = compute_importance_gbt(
      mus_train, ys_train, mus_test, ys_test)
  assert importance_matrix.shape[0] == mus_train.shape[0]
  assert importance_matrix.shape[1] == ys_train.shape[0]
  scores["informativeness_train"] = train_err
  scores["informativeness_test"] = test_err
  scores["disentanglement"] = disentanglement(importance_matrix)
  scores["completeness"] = completeness(importance_matrix)
  return scores


@gin.configurable(
    "dci_validation",
    blacklist=["observations", "labels", "representation_function"])
def compute_dci_on_fixed_data(observations, labels, representation_function,
                              train_percentage=gin.REQUIRED, batch_size=100):
  """Computes the DCI scores on the fixed set of observations and labels.

  Args:
    observations: Observations on which to compute the score. Observations have
      shape (num_observations, 64, 64, num_channels).
    labels: Observed factors of variations.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    train_percentage: Percentage of observations used for training.
    batch_size: Batch size used to compute the representation.

  Returns:
    DCI score.
  """
  mus = utils.obtain_representation(observations, representation_function,
                                    batch_size)
  assert labels.shape[1] == observations.shape[0], "Wrong labels shape."
  assert mus.shape[1] == observations.shape[0], "Wrong representation shape."
  mus_train, mus_test = utils.split_train_test(
      mus,
      train_percentage)
  ys_train, ys_test = utils.split_train_test(
      labels,
      train_percentage)
  return _compute_dci(mus_train, ys_train, mus_test, ys_test)


def compute_importance_gbt(x_train, y_train, x_test, y_test):
  """Compute importance based on gradient boosted trees."""
  num_factors = y_train.shape[0]
  num_codes = x_train.shape[0]
  importance_matrix = np.zeros(shape=[num_codes, num_factors],
                               dtype=np.float64)
  train_loss = []
  test_loss = []
  for i in range(num_factors):
    model = ensemble.GradientBoostingClassifier()
    model.fit(x_train.T, y_train[i, :])
    importance_matrix[:, i] = np.abs(model.feature_importances_)
    train_loss.append(np.mean(model.predict(x_train.T) == y_train[i, :]))
    test_loss.append(np.mean(model.predict(x_test.T) == y_test[i, :]))
  return importance_matrix, np.mean(train_loss), np.mean(test_loss)


def disentanglement_per_code(importance_matrix):
  """Compute disentanglement score of each code."""
  # importance_matrix is of shape [num_codes, num_factors].
  return 1. - scipy.stats.entropy(importance_matrix.T + 1e-11,
                                  base=importance_matrix.shape[1])


def disentanglement(importance_matrix):
  """Compute the disentanglement score of the representation."""
  per_code = disentanglement_per_code(importance_matrix)
  if importance_matrix.sum() == 0.:
    importance_matrix = np.ones_like(importance_matrix)
  code_importance = importance_matrix.sum(axis=1) / importance_matrix.sum()

  return np.sum(per_code*code_importance)


def completeness_per_factor(importance_matrix):
  """Compute completeness of each factor."""
  # importance_matrix is of shape [num_codes, num_factors].
  return 1. - scipy.stats.entropy(importance_matrix + 1e-11,
                                  base=importance_matrix.shape[0])


def completeness(importance_matrix):
  """"Compute completeness of the representation."""
  per_factor = completeness_per_factor(importance_matrix)
  if importance_matrix.sum() == 0.:
    importance_matrix = np.ones_like(importance_matrix)
  factor_importance = importance_matrix.sum(axis=0) / importance_matrix.sum()
  return np.sum(per_factor*factor_importance)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for relational_layers.py."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from disentanglement_lib.evaluation.abstract_reasoning import relational_layers
import numpy as np
import tensorflow.compat.v1 as tf


def _create_positional_encoding_matrices():
  """Shared input/output pair for the positional encoding tests."""
  input_array = np.arange(24, dtype=np.float64).reshape((1, 4, 3, 2))
  output_array = np.eye(4)
  output_array = np.repeat(np.expand_dims(output_array, -1), 2, axis=-1)
  output_array = np.expand_dims(output_array, 0)
  return input_array, output_array


class RelationalLayersTest(tf.test.TestCase):

  def test_repeat_for_tensor(self):
    a = np.arange(24).reshape((1, 4, 3, 2))
    shouldbe = np.concatenate([a] * 3, axis=-2)
    result = self.evaluate(relational_layers.repeat(tf.constant(a), 3, axis=-2))
    self.assertAllClose(shouldbe, result)

  def test_pairwise_edge_embeddings_for_tensor(self):
    a = np.array([[[1], [2]]])
    shouldbe = np.array([[[[1, 1], [1, 2]], [[2, 1], [2, 2]]]])
    layer = relational_layers.PairwiseEdgeEmbeddings()
    result = self.evaluate(layer(tf.constant(a)))
    self.assertAllClose(shouldbe, result)

  def test_relational_layer_for_tensor(self):
    a = np.array([[[1], [2]]])
    shouldbe = np.array([[[2, 3], [4, 3]]])
    layer = relational_layers.RelationalLayer(
        tf.keras.layers.Lambda(lambda x: x),
        tf.keras.layers.Lambda(lambda x: tf.reduce_sum(x, axis=-2)))
    result = self.evaluate(layer(tf.constant(a)))
    self.assertAllClose(shouldbe, result)

  def test_positional_encoding_like_for_static_shape_tensor(self):
    value, shouldbe = _create_positional_encoding_matrices()
    a = tf.constant(value)
    output_tensor = relational_layers.positional_encoding_like(a, -3, -2)
    result = self.evaluate(output_tensor)
    self.assertEqual((1, 4, 4, 2), result.shape)
    self.assertAllClose(shouldbe, result)

  def test_positional_encoding_like_for_dynamic_shape_tensor(self):
    value, shouldbe = _create_positional_encoding_matrices()
    a = tf.placeholder(tf.float32, shape=(None, 4, 3, 2))
    output_tensor = relational_layers.positional_encoding_like(a, -3, -2)
    # Check the static shape.
    self.assertEqual([None, 4, 4, 2], output_tensor.get_shape().as_list())
    # Check the solution.
    with self.session() as sess:
      result = sess.run(output_tensor, feed_dict={a: value})
    self.assertAllClose(shouldbe, result)

  def test_add_positional_encoding_layer_for_tensor(self):
    value, shouldbe_positional = _create_positional_encoding_matrices()
    shouldbe = np.concatenate([value, shouldbe_positional], axis=-2)
    a = tf.constant(value)
    output_tensor = relational_layers.AddPositionalEncoding(-3, -2)(a)
    result = self.evaluate(output_tensor)
    self.assertAllClose(shouldbe, result)

  def test_stack_answers_for_tensors(self):
    # Tensors used for testing.
    context = np.arange(24).reshape((2, 3, 4))
    answers = np.arange(24, 48).reshape((2, 3, 4))
    # Compute the correct solutions.
    results = []
    for i in range(answers.shape[-1]):
      results.append(
          np.concatenate([context, answers[:, :, i:(i + 1)]], axis=-1))
    shouldbe = np.stack(results, axis=-2)
    # Compute the solution based on the layer.
    layer = relational_layers.StackAnswers(answer_axis=-1, stack_axis=-2)
    result = self.evaluate(layer([tf.constant(context), tf.constant(answers)]))
    # Check that they are the same.
    self.assertAllClose(shouldbe, result)

  def test_multi_dim_batch_apply_for_tensors(self):
    # Tensors used for testing.
    input_tensor = np.arange(24).reshape((2, 3, 4))
    kernel = np.arange(24, 36).reshape((4, 3))
    # Compute the correct solutions.
    shouldbe = np.matmul(input_tensor, kernel)
    # Compute the solution based on the layer.
    layer = relational_layers.MultiDimBatchApply(
        tf.keras.layers.Lambda(lambda x: tf.matmul(x, tf.constant(kernel))),
        num_dims_to_keep=1)
    result = self.evaluate(layer(tf.constant(input_tensor)))
    # Check that they are the same.
    self.assertAllClose(shouldbe, result)


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides named, gin configurable ground truth data sets."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.data.ground_truth import cars3d
from disentanglement_lib.data.ground_truth import dsprites
from disentanglement_lib.data.ground_truth import dummy_data
from disentanglement_lib.data.ground_truth import mpi3d
from disentanglement_lib.data.ground_truth import norb
from disentanglement_lib.data.ground_truth import shapes3d
import gin.tf


@gin.configurable("dataset")
def get_named_ground_truth_data(name):
  """Returns ground truth data set based on name.

  Args:
    name: String with the name of the dataset.

  Raises:
    ValueError: if an invalid data set name is provided.
  """

  if name == "dsprites_full":
    return dsprites.DSprites([1, 2, 3, 4, 5])
  elif name == "dsprites_noshape":
    return dsprites.DSprites([2, 3, 4, 5])
  elif name == "color_dsprites":
    return dsprites.ColorDSprites([1, 2, 3, 4, 5])
  elif name == "noisy_dsprites":
    return dsprites.NoisyDSprites([1, 2, 3, 4, 5])
  elif name == "scream_dsprites":
    return dsprites.ScreamDSprites([1, 2, 3, 4, 5])
  elif name == "smallnorb":
    return norb.SmallNORB()
  elif name == "cars3d":
    return cars3d.Cars3D()
  elif name == "mpi3d_toy":
    return mpi3d.MPI3D(mode="mpi3d_toy")
  elif name == "mpi3d_realistic":
    return mpi3d.MPI3D(mode="mpi3d_realistic")
  elif name == "mpi3d_real":
    return mpi3d.MPI3D(mode="mpi3d_real")
  elif name == "shapes3d":
    return shapes3d.Shapes3D()
  elif name == "dummy_data":
    return dummy_data.DummyData()
  else:
    raise ValueError("Invalid data set name.")



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Different studies that can be reproduced."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from disentanglement_lib.config.abstract_reasoning_study_v1.stage1 import sweep as abstract_reasoning_study_v1
from disentanglement_lib.config.fairness_study_v1 import sweep as fairness_study_v1
from disentanglement_lib.config.tests import sweep as tests
from disentanglement_lib.config.unsupervised_study_v1 import sweep as unsupervised_study_v1

STUDIES = {
    "unsupervised_study_v1": unsupervised_study_v1.UnsupervisedStudyV1(),
    "abstract_reasoning_study_v1":
        abstract_reasoning_study_v1.AbstractReasoningStudyV1(),
    "fairness_study_v1":
        fairness_study_v1.FairnessStudyV1(),
    "test": tests.TestStudy(),
}



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the visualize_scores.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized
from disentanglement_lib.visualize import visualize_scores
import numpy as np


class VisualizeScoresTest(parameterized.TestCase):

  @parameterized.parameters((np.zeros((5, 10), dtype=np.float32),
                             np.zeros((2, 10), dtype=np.float32),
                             np.zeros((10, 10), dtype=np.float32)))
  def plot_recovery_vs_independent(self, matrix):
    visualize_scores.plot_recovery_vs_independent(
        matrix, self.create_tempdir().full_path, "save_image.png")


class ComputeMatrixStatsTest(parameterized.TestCase):

  @parameterized.parameters(
      (np.eye(5), [(0, 0)], [1, 0, 0, 0, 0], [1, 0, 0, 0, 0], 2),
      (np.triu(np.ones(5)), [(0, 0)], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1], 10),
      (np.ones((5, 5)), [(0, 0)], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1], 10),
      (np.array([[1, 1, 0, 0, 0],
                 [0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0],
                 [0, 0, 0, 1, 1]]),
       [(0, 0)], [1, 0, 0, 0, 0], [1, 1, 0, 0, 0], 3),
      (np.array([[1, 1, 0, 0, 0],
                 [0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 0],
                 [0, 1, 0, 0, 0],
                 [0, 0, 0, 1, 1]]),
       [(0, 0)], [1, 0, 0, 1, 0], [1, 1, 0, 0, 0], 4)
  )
  def test_bfs(self, matrix, to_visit, visited_factors, visited_codes,
               visited_size):
    factors = np.zeros(matrix.shape[0])
    codes = np.zeros(matrix.shape[1])
    factors, codes, size = visualize_scores.bfs(
        matrix, to_visit, factors, codes, 1)

    self.assertEqual((factors == visited_factors).all(), True)
    self.assertEqual((codes == visited_codes).all(), True)
    self.assertEqual(size, visited_size)

  @parameterized.parameters(
      (np.zeros((5, 5)), 0),
      (np.eye(5), 5),
      (np.triu(np.ones(5)), 1),
      (np.ones((5, 5)), 1)
  )
  def test_precision(self, matrix, cc):
    cc_count = visualize_scores.precision(matrix, 1)
    self.assertEqual(cc, cc_count)

  @parameterized.parameters(
      (np.zeros((5, 5)), 0),
      (np.eye(5), 5),
      (np.triu(np.ones(5)), 5),
      (np.ones((5, 5)), 5)
  )
  def test_recall(self, matrix, cc):
    cc_count = visualize_scores.recall(matrix, 1)
    self.assertEqual(cc, cc_count)

if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test study."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.config import study
from disentanglement_lib.utils import resources


class TestStudy(study.Study):
  """Defines a study for testing."""

  def get_model_config(self, model_num=0):
    """Returns model bindings and config file."""
    return [], resources.get_file(
        "config/tests/methods/unsupervised/train_test.gin")

  def get_postprocess_config_files(self):
    """Returns postprocessing config files."""
    return list(
        resources.get_files_in_folder(
            "config/tests/postprocessing/postprocess_test_configs"))

  def get_eval_config_files(self):
    """Returns evaluation config files."""
    return list(
        resources.get_files_in_folder(
            "config/tests/evaluation/evaluate_test_configs"))



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to visualize the unified scores.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os

import matplotlib
matplotlib.use("Agg")  # Set headless-friendly backend.
import matplotlib.pyplot as plt  # pylint: disable=g-import-not-at-top
import numpy as np
import pandas as pd
import seaborn as sns
import tensorflow.compat.v1 as tf


def heat_square(matrix, output_dir, name, xlabel, ylabel, max_val=None,
                factor_names=None):
  """Plot values of a matrix.

  Each entry is represented as a square of increasing size and different color.

  Args:
    matrix: Matrix of values to plot. Values should be in range [0, max_val].
    output_dir: Where to save the image.
    name: File name.
    xlabel: Name of the x axis of the matrix.
    ylabel: Name of the y axis of the matrix.
    max_val: Maximum value acceptable in the matrix. If None, the max_val will
      be set as the maximum value in the matrix.
    factor_names: Names of the factors of variation.
  """
  sns.set_context("notebook", font_scale=1.5, rc={"lines.linewidth": 2})
  sns.set_style("whitegrid")
  fig, _ = plt.subplots()
  plot_grid = plt.GridSpec(1, 15, hspace=0.2, wspace=1.2)
  ax = plt.subplot(plot_grid[:, :-1])
  if max_val is None:
    max_val = np.max(matrix)
    if max_val == 0:
      max_val = 1.
  else:
    if max_val < np.max(matrix):
      raise ValueError("The matrix has maximum value larger than max_val")
  palette = sns.color_palette("Blues", 256)
  # Estimates the area of the squares: the length of the edge is
  # roughly: length of the grid in inches * how many points per inch - space for
  # the axis names times * 14/15 as the last 1/15 part of the figure is occupied
  # by the colorbar legend.
  size_scale = ((((ax.get_position().xmax - ax.get_position().xmin) *
                  fig.get_size_inches()[0] * fig.get_dpi() - 40) * 14/15*0.8) /
                (matrix.shape[0]))**2
  plot_matrix_squares(matrix, max_val, palette, size_scale, ax)
  plt.xticks(range(matrix.shape[0]))
  if factor_names is not None:
    plt.yticks(range(matrix.shape[1]), factor_names)
  else:
    plt.yticks(range(matrix.shape[1]))
  plt.xlabel(xlabel)
  plt.ylabel(ylabel)
  # Add color legend on the right side of the plot.
  ax = plt.subplot(plot_grid[:, -1])
  plot_bar_palette(palette, max_val, ax)

  if not tf.gfile.IsDirectory(output_dir):
    tf.gfile.MakeDirs(output_dir)
  output_path = os.path.join(output_dir, "{}.png".format(name))
  with tf.gfile.Open(output_path, "wb") as path:
    fig.savefig(path, bbox_inches="tight")


def plot_matrix_squares(matrix, max_val, palette, size_scale, ax):
  """Grid of squares where the size is proportional to the matrix values.

  Args:
    matrix: Matrix of values to plot.
    max_val: Maximum value that is allowed in the matrix.
    palette: Color palette.
    size_scale: Maximum size of the squares.
    ax: Axis of the subplot.
  """
  tmp = pd.melt(pd.DataFrame(matrix).reset_index(), id_vars="index")
  # The columns of the dataframe are: index, variable and value.
  def to_color(val):
    ind = int(val/max_val*255)
    return palette[ind]
  ax.scatter(x=tmp["index"], y=tmp["variable"],
             s=size_scale * tmp["value"]/max_val, marker="s",
             c=tmp["value"].apply(to_color))
  ax.set_xticks([v+0.5 for v in range(matrix.shape[0])], minor=True)
  ax.set_yticks([v+0.5 for v in range(matrix.shape[1])], minor=True)

  ax.grid(False, "major")
  ax.grid(True, "minor")
  ax.set_xlim([-0.5, matrix.shape[0] - 0.5])
  ax.set_ylim([-0.5, matrix.shape[1] - 0.5])
  ax.tick_params(right=False, top=False, left=False, bottom=False)
  ax.set_aspect(aspect=1.)


def plot_bar_palette(palette, max_val, ax):
  """Plot color bar legend."""
  col_x = [0]*len(palette)
  bar_y = np.linspace(0, max_val, 256, ax)

  bar_height = bar_y[1] - bar_y[0]
  ax.barh(bar_y, np.array([5]*len(palette)), height=bar_height, left=col_x,
          align="center", color=palette, linewidth=0)

  ax.set_xlim(1, 2)
  ax.set_ylim(0, max_val)
  ax.grid(False)
  ax.set_xticks([])
  ax.set_yticks(np.linspace(0, max_val, 3))
  ax.yaxis.tick_right()


def plot_recovery_vs_independent(matrix, output_dir, name):
  """Plot how many factors are recovered and in how many independent groups.

  Plot how many factors of variation are independently captured in a
  representation at different thresholds. It takes as input a matrix
  relating factors of variation and latent dimensions, sort the elements and
  then plot for each threshold (1) how many factors are discovered and (2)
  how many factors are encoded independently in the representation.

  Args:
    matrix: Contains statistical relations between factors of variation and
      latent codes.
    output_dir: Output directory where to save the plot.
    name: Filename of the plot.
  """
  thresholds = np.sort(matrix.flatten())[::-1]
  precisions = [precision(matrix, x) for x in thresholds]
  recalls = [recall(matrix, x) for x in thresholds]
  sns.set_context("notebook", font_scale=1.5, rc={"lines.linewidth": 2})
  sns.set_style("whitegrid")
  fig, ax = plt.subplots()
  palette = sns.color_palette()
  plt.plot(range(thresholds.shape[0]), precisions, label="Independent groups",
           color=palette[0], linewidth=3)
  plt.plot(range(thresholds.shape[0]), recalls, "--", label="Discovered",
           color=palette[1], linewidth=3)
  thresholds_ids = range(0, thresholds.shape[0], 10)
  plt.xticks(thresholds_ids, np.around(thresholds[thresholds_ids], 2))
  ax.set_ylim([0, matrix.shape[0]*1.1])
  ax.tick_params(right=False, top=False, left=False, bottom=False)
  ax.set_yticks(np.linspace(0, matrix.shape[0], matrix.shape[0]+1))
  plt.legend(loc="upper center", bbox_to_anchor=(0.5, 1.25), ncol=2)
  plt.xlabel("Threshold")
  plt.ylabel("Number of Factors")
  if not tf.gfile.IsDirectory(output_dir):
    tf.gfile.MakeDirs(output_dir)
  output_path = os.path.join(output_dir, name+".png")
  with tf.gfile.Open(output_path, "wb") as path:
    fig.savefig(path, bbox_inches="tight")


def precision(matrix, th):
  """How many independent components are discovered for a given threshold.

  Args:
    matrix: Adjacency matrix  of shape (num_codes, num_factors) encoding the
      statistical relations between factors and codes.
    th: Eliminate all edges smaller than this threshold.

  Returns:
    Number of connected components.
  """
  tmp = matrix.copy()
  tmp[tmp < th] = 0
  factors = np.zeros(tmp.shape[0])
  codes = np.zeros(tmp.shape[1])
  cc = 0

  for i in range(len(factors)):
    if factors[i] == 0:
      to_visit = [(i, 0)]
      factors, codes, size = bfs(tmp, to_visit, factors, codes, 1)
      if size > 1:
        cc += 1
  return cc


def recall(matrix, th):
  """How many factors are discovered for a given threshold.

  Counts as many factors of variation are captured in the representation.
  First, we remove all edges in the adjacency matrix with weight smaller than
  the threshold. Then, we count how many factors are connected to some codes.

  Args:
    matrix: Adjacency matrix for the graph.
    th: Eliminate all edges smaller than this threshold.

  Returns:
    Number of discovered factors of variation for the given threshold.
  """
  tmp = matrix.copy()
  tmp[tmp < th] = 0
  return np.sum(np.sum(tmp, axis=1) != 0)


def bfs(matrix, to_visit, factors, codes, size):
  """Traverse the matrix across connected components.

  Implements breadth first search on an adjacency matrix. In our case, the
  adjacency matrix encodes the statistical relations between factors of
  variation and codes. This is used to traverse the adjacency matrix and
  discover whether a factor is captured in multiple codes and whether there is a
  path in the graph connecting two factors.

  Args:
    matrix: Adjacency matrix for the graph.
    to_visit: Queue with the nodes to visit. We index the factors and codes in
      the adjacency matrix and implement the queue with an array containing the
      nodes that need to  be visited.
    factors: Array of shape (num_factors, ) with flags marking whether factors
      of variation are visited.
    codes: Array of shape (num_codes, ) with flags marking whether codes are
      visited.
    size: Count how many node are in the same connected component.

  Returns:
    factors: Array of shape (num_factors, ) with flags marking whether factors
      of variation are visited.
    codes: Array of shape (num_codes, ) with flags marking whether codes are
      visited.
    size: How many nodes were visited.
  """
  (current_node, flag) = to_visit.pop()
  if flag == 0:
    factors[current_node] = 1
    for i in range(len(matrix[current_node, :])):
      if matrix[current_node, i] != 0:
        if codes[i] == 0:
          to_visit.append((i, 1))
          size += 1
          factors, codes, size = bfs(matrix, to_visit, factors, codes, size)
  else:
    codes[current_node] = 1
    for i in range(len(matrix[:, current_node])):
      if matrix[i, current_node] != 0:
        if factors[i] == 0:
          to_visit.append((i, 0))
          size += 1
          factors, codes, size = bfs(matrix, to_visit, factors, codes, size)
  return factors, codes, size



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for modularity_explicitness.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import absltest
from disentanglement_lib.data.ground_truth import dummy_data
from disentanglement_lib.evaluation.metrics import modularity_explicitness
import numpy as np
from six.moves import range
import gin.tf


def _identity_discretizer(target, num_bins):
  del num_bins
  return target


class ModularityTest(absltest.TestCase):

  def test_diagonal(self):
    importance_matrix = np.diag(5.*np.ones(5))
    result = modularity_explicitness.modularity(importance_matrix)
    np.testing.assert_allclose(result, 1.0)

  def test_diagonal_empty_codes(self):
    importance_matrix = np.array([[1., 0.,], [0., 1.], [0., 0.]])
    result = modularity_explicitness.modularity(importance_matrix)
    np.testing.assert_allclose(result, 2./3.)

  def test_zero(self):
    importance_matrix = np.zeros(shape=[10, 10], dtype=np.float64)
    result = modularity_explicitness.modularity(importance_matrix)
    np.testing.assert_allclose(result, .0)

  def test_redundant_codes(self):
    importance_matrix = np.diag(5.*np.ones(5))
    importance_matrix = np.vstack([importance_matrix, importance_matrix])
    result = modularity_explicitness.modularity(importance_matrix)
    np.testing.assert_allclose(result, 1.)

  def test_missed_factors(self):
    importance_matrix = np.diag(5.*np.ones(5))
    result = modularity_explicitness.modularity(importance_matrix[:2, :])
    np.testing.assert_allclose(result, 1.0)

  def test_one_code_two_factors(self):
    importance_matrix = np.diag(5.*np.ones(5))
    importance_matrix = np.hstack([importance_matrix, importance_matrix])
    result = modularity_explicitness.modularity(importance_matrix)
    np.testing.assert_allclose(result, 1. - 1./9)


class ModularityExplicitnessTest(absltest.TestCase):

  def test_metric(self):
    gin.bind_parameter("discretizer.discretizer_fn", _identity_discretizer)
    gin.bind_parameter("discretizer.num_bins", 10)
    ground_truth_data = dummy_data.IdentityObservationsData()
    representation_function = lambda x: np.array(x, dtype=np.float64)
    random_state = np.random.RandomState(0)
    scores = modularity_explicitness.compute_modularity_explicitness(
        ground_truth_data, representation_function, random_state, None, 3000,
        3000)
    self.assertBetween(scores["modularity_score"], 0.9, 1.0)

  def test_bad_metric(self):
    gin.bind_parameter("discretizer.discretizer_fn", _identity_discretizer)
    gin.bind_parameter("discretizer.num_bins", 10)
    ground_truth_data = dummy_data.IdentityObservationsData()
    random_state_rep = np.random.RandomState(0)
    # The representation which randomly permutes the factors, should have equal
    # non-zero MI which should give a low modularity score.
    def representation_function(x):
      code = np.array(x, dtype=np.float64)
      for i in range(code.shape[0]):
        code[i, :] = random_state_rep.permutation(code[i, :])
      return code
    random_state = np.random.RandomState(0)
    scores = modularity_explicitness.compute_modularity_explicitness(
        ground_truth_data, representation_function, random_state, None, 20000,
        20000)
    self.assertBetween(scores["modularity_score"], 0.0, 0.2)

  def test_duplicated_latent_space(self):
    gin.bind_parameter("discretizer.discretizer_fn", _identity_discretizer)
    gin.bind_parameter("discretizer.num_bins", 10)
    ground_truth_data = dummy_data.IdentityObservationsData()
    def representation_function(x):
      x = np.array(x, dtype=np.float64)
      return np.hstack([x, x])
    random_state = np.random.RandomState(0)
    scores = modularity_explicitness.compute_modularity_explicitness(
        ground_truth_data, representation_function, random_state, None, 3000,
        3000)
    self.assertBetween(scores["modularity_score"], 0.9, 1.0)

if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for unsupervised_metrics.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import absltest
from disentanglement_lib.evaluation.metrics import unsupervised_metrics
import numpy as np
import scipy


class UnsupervisedMetricsTest(absltest.TestCase):

  def test_gaussian_total_correlation_zero(self):
    score = unsupervised_metrics.gaussian_total_correlation(
        np.diag(np.ones(5, dtype=np.float64)))
    self.assertBetween(score, -0.01, 0.01)

  def test_gaussian_total_correlation_same(self):
    """Check that the results of the both functions are the same."""
    cov = np.array([[1, 0.9], [0.9, 1.0]], dtype=np.float32)
    means = np.array([0.0, 0.0], dtype=np.float32)
    cov_central = np.diag(np.diag(cov))
    shouldbe = unsupervised_metrics.kl_gaussians_numerically_unstable(
        means, cov, means, cov_central, 2)
    score = unsupervised_metrics.gaussian_total_correlation(cov)
    self.assertBetween(score, shouldbe - 0.01, shouldbe + 0.01)

  def test_gaussian_wasserstein_correlation_zero(self):
    score = unsupervised_metrics.gaussian_wasserstein_correlation(
        np.diag(np.ones(5, dtype=np.float64)))
    self.assertBetween(score, -0.01, 0.01)

  def test_gaussian_wasserstein_correlation_same(self):
    cov = np.array([[1, 0.9], [0.9, 1.0]], dtype=np.float32)
    score = unsupervised_metrics.gaussian_wasserstein_correlation(cov)
    cov_only_diagonal = np.diag(np.diag(cov))
    sqrtm = scipy.linalg.sqrtm(np.matmul(cov, cov_only_diagonal))
    shouldbe = np.trace(cov + cov_only_diagonal - 2 * sqrtm)
    self.assertBetween(score, shouldbe - 0.01, shouldbe + 0.01)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluation protocol to compute metrics."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import inspect
import os
import time
import warnings

from disentanglement_lib.data.ground_truth import named_data
from disentanglement_lib.evaluation.metrics import beta_vae  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import dci  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import downstream_task  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import factor_vae  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import fairness  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import irs  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import mig  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import modularity_explicitness  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import reduced_downstream_task  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import sap_score  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import strong_downstream_task  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import unified_scores  # pylint: disable=unused-import
from disentanglement_lib.evaluation.metrics import unsupervised_metrics  # pylint: disable=unused-import
from disentanglement_lib.utils import results
import numpy as np
import tensorflow.compat.v1 as tf
import tensorflow_hub as hub

import gin.tf


def evaluate_with_gin(model_dir,
                      output_dir,
                      overwrite=False,
                      gin_config_files=None,
                      gin_bindings=None):
  """Evaluate a representation based on the provided gin configuration.

  This function will set the provided gin bindings, call the evaluate()
  function and clear the gin config. Please see the evaluate() for required
  gin bindings.

  Args:
    model_dir: String with path to directory where the representation is saved.
    output_dir: String with the path where the evaluation should be saved.
    overwrite: Boolean indicating whether to overwrite output directory.
    gin_config_files: List of gin config files to load.
    gin_bindings: List of gin bindings to use.
  """
  if gin_config_files is None:
    gin_config_files = []
  if gin_bindings is None:
    gin_bindings = []
  gin.parse_config_files_and_bindings(gin_config_files, gin_bindings)
  evaluate(model_dir, output_dir, overwrite)
  gin.clear_config()


@gin.configurable(
    "evaluation", blacklist=["model_dir", "output_dir", "overwrite"])
def evaluate(model_dir,
             output_dir,
             overwrite=False,
             evaluation_fn=gin.REQUIRED,
             random_seed=gin.REQUIRED,
             name=""):
  """Loads a representation TFHub module and computes disentanglement metrics.

  Args:
    model_dir: String with path to directory where the representation function
      is saved.
    output_dir: String with the path where the results should be saved.
    overwrite: Boolean indicating whether to overwrite output directory.
    evaluation_fn: Function used to evaluate the representation (see metrics/
      for examples).
    random_seed: Integer with random seed used for training.
    name: Optional string with name of the metric (can be used to name metrics).
  """
  # Delete the output directory if it already exists.
  if tf.gfile.IsDirectory(output_dir):
    if overwrite:
      tf.gfile.DeleteRecursively(output_dir)
    else:
      raise ValueError("Directory already exists and overwrite is False.")

  # Set up time to keep track of elapsed time in results.
  experiment_timer = time.time()

  # Automatically set the proper data set if necessary. We replace the active
  # gin config as this will lead to a valid gin config file where the data set
  # is present.
  if gin.query_parameter("dataset.name") == "auto":
    # Obtain the dataset name from the gin config of the previous step.
    gin_config_file = os.path.join(model_dir, "results", "gin",
                                   "postprocess.gin")
    gin_dict = results.gin_dict(gin_config_file)
    with gin.unlock_config():
      gin.bind_parameter("dataset.name", gin_dict["dataset.name"].replace(
          "'", ""))
  dataset = named_data.get_named_ground_truth_data()

  # Path to TFHub module of previously trained representation.
  module_path = os.path.join(model_dir, "tfhub")
  with hub.eval_function_for_module(module_path) as f:

    def _representation_function(x):
      """Computes representation vector for input images."""
      output = f(dict(images=x), signature="representation", as_dict=True)
      return np.array(output["default"])

    # Computes scores of the representation based on the evaluation_fn.
    if _has_kwarg_or_kwargs(evaluation_fn, "artifact_dir"):
      artifact_dir = os.path.join(model_dir, "artifacts", name)
      results_dict = evaluation_fn(
          dataset,
          _representation_function,
          random_state=np.random.RandomState(random_seed),
          artifact_dir=artifact_dir)
    else:
      # Legacy code path to allow for old evaluation metrics.
      warnings.warn(
          "Evaluation function does not appear to accept an"
          " `artifact_dir` argument. This may not be compatible with "
          "future versions.", DeprecationWarning)
      results_dict = evaluation_fn(
          dataset,
          _representation_function,
          random_state=np.random.RandomState(random_seed))

  # Save the results (and all previous results in the pipeline) on disk.
  original_results_dir = os.path.join(model_dir, "results")
  results_dir = os.path.join(output_dir, "results")
  results_dict["elapsed_time"] = time.time() - experiment_timer
  results.update_result_directory(results_dir, "evaluation", results_dict,
                                  original_results_dir)


def _has_kwarg_or_kwargs(f, kwarg):
  """Checks if the function has the provided kwarg or **kwargs."""
  # For gin wrapped functions, we need to consider the wrapped function.
  if hasattr(f, "__wrapped__"):
    f = f.__wrapped__
  (args, _, kwargs, _, _, _, _) = inspect.getfullargspec(f)
  if kwarg in args or kwargs is not None:
    return True
  return False



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for visualize_util.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
from absl.testing import absltest
from disentanglement_lib.visualize import visualize_util
import numpy as np


class VisualizeUtilTest(absltest.TestCase):

  def test_save_image(self):
    image = np.zeros((128, 256, 3), dtype=np.float32)
    path = os.path.join(self.create_tempdir().full_path, "save_image.png")
    visualize_util.save_image(image, path)

  def test_save_image_grayscale(self):
    image = np.ones((128, 256, 1), dtype=np.float32)
    path = os.path.join(self.create_tempdir().full_path,
                        "save_image_grayscale.png")
    visualize_util.save_image(image, path)

  def test_grid_save_images(self):
    images = np.zeros((18, 128, 256, 3), dtype=np.float32)
    path = os.path.join(self.create_tempdir().full_path, "grid_save_images.png")
    visualize_util.grid_save_images(images, path)

  def test_save_animation(self):
    path = os.path.join(self.create_tempdir().full_path, "animation.gif")
    images = np.ones((18, 128, 256, 3), dtype=np.float32)
    visualize_util.save_animation([images, images], path, fps=18)

  def cycle_factor(self):
    result = list(visualize_util.cycle_factor(1, 3, 4))
    shouldbe = [1, 2, 2, 1, 0, 0]
    self.assertAllEqual(result, shouldbe)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SmallNORB dataset."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
from disentanglement_lib.data.ground_truth import ground_truth_data
from disentanglement_lib.data.ground_truth import util
import numpy as np
import PIL
from six.moves import range
from six.moves import zip
import tensorflow.compat.v1 as tf


SMALLNORB_TEMPLATE = os.path.join(
    os.environ.get("DISENTANGLEMENT_LIB_DATA", "."), "small_norb",
    "smallnorb-{}-{}.mat")

SMALLNORB_CHUNKS = [
    "5x46789x9x18x6x2x96x96-training",
    "5x01235x9x18x6x2x96x96-testing",
]


class SmallNORB(ground_truth_data.GroundTruthData):
  """SmallNORB dataset.

  The data set can be downloaded from
  https://cs.nyu.edu/~ylclab/data/norb-v1.0-small/. Images are resized to 64x64.

  The ground-truth factors of variation are:
  0 - category (5 different values)
  1 - elevation (9 different values)
  2 - azimuth (18 different values)
  3 - lighting condition (6 different values)

  The instance in each category is randomly sampled when generating the images.
  """

  def __init__(self):
    self.images, features = _load_small_norb_chunks(SMALLNORB_TEMPLATE,
                                                    SMALLNORB_CHUNKS)
    self.factor_sizes = [5, 10, 9, 18, 6]
    # Instances are not part of the latent space.
    self.latent_factor_indices = [0, 2, 3, 4]
    self.num_total_factors = features.shape[1]
    self.index = util.StateSpaceAtomIndex(self.factor_sizes, features)
    self.state_space = util.SplitDiscreteStateSpace(self.factor_sizes,
                                                    self.latent_factor_indices)

  @property
  def num_factors(self):
    return self.state_space.num_latent_factors

  @property
  def factors_num_values(self):
    return [self.factor_sizes[i] for i in self.latent_factor_indices]

  @property
  def observation_shape(self):
    return [64, 64, 1]


  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    return self.state_space.sample_latent_factors(num, random_state)

  def sample_observations_from_factors(self, factors, random_state):
    all_factors = self.state_space.sample_all_factors(factors, random_state)
    indices = self.index.features_to_index(all_factors)
    return np.expand_dims(self.images[indices].astype(np.float32), axis=3)


def _load_small_norb_chunks(path_template, chunk_names):
  """Loads several chunks of the small norb data set for final use."""
  list_of_images, list_of_features = _load_chunks(path_template, chunk_names)
  features = np.concatenate(list_of_features, axis=0)
  features[:, 3] = features[:, 3] / 2  # azimuth values are 0, 2, 4, ..., 24
  return np.concatenate(list_of_images, axis=0), features




def _load_chunks(path_template, chunk_names):
  """Loads several chunks of the small norb data set into lists."""
  list_of_images = []
  list_of_features = []
  for chunk_name in chunk_names:
    norb = _read_binary_matrix(path_template.format(chunk_name, "dat"))
    list_of_images.append(_resize_images(norb[:, 0]))
    norb_class = _read_binary_matrix(path_template.format(chunk_name, "cat"))
    norb_info = _read_binary_matrix(path_template.format(chunk_name, "info"))
    list_of_features.append(np.column_stack((norb_class, norb_info)))
  return list_of_images, list_of_features


def _read_binary_matrix(filename):
  """Reads and returns binary formatted matrix stored in filename."""
  with tf.gfile.GFile(filename, "rb") as f:
    s = f.read()
    magic = int(np.frombuffer(s, "int32", 1))
    ndim = int(np.frombuffer(s, "int32", 1, 4))
    eff_dim = max(3, ndim)
    raw_dims = np.frombuffer(s, "int32", eff_dim, 8)
    dims = []
    for i in range(0, ndim):
      dims.append(raw_dims[i])

    dtype_map = {
        507333717: "int8",
        507333716: "int32",
        507333713: "float",
        507333715: "double"
    }
    data = np.frombuffer(s, dtype_map[magic], offset=8 + eff_dim * 4)
  data = data.reshape(tuple(dims))
  return data


def _resize_images(integer_images):
  resized_images = np.zeros((integer_images.shape[0], 64, 64))
  for i in range(integer_images.shape[0]):
    image = PIL.Image.fromarray(integer_images[i, :, :])
    image = image.resize((64, 64), PIL.Image.ANTIALIAS)
    resized_images[i, :, :] = image
  return resized_images / 255.



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Keras models to perform abstract reasoning."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from disentanglement_lib.evaluation.abstract_reasoning import relational_layers
import gin
import tensorflow.compat.v1 as tf
import tensorflow_hub as hub
from tensorflow.contrib import tpu as contrib_tpu


@gin.configurable
class TwoStageModel(object):
  """Two stage model for abstract reasoning tasks.

  This class implements a flexible variation of the Wild Relation Networks model
  of Barrett et al., 2018 (https://arxiv.org/abs/1807.04225). There are two
  stages:

  1. Embedding: This embeds the patches of the PGM each indepently into a lower
    dimensional embedding (e.g., via CNN). It is also supported to take one-hot
    embeddings or integer embeddings of the ground-truth factors (as baselines).

  2. Reasoning: This performs reasoning on the embeddings of the patches of the
    PGM and returns the solution.
  """

  def __init__(self,
               embedding_model_class=gin.REQUIRED,
               reasoning_model_class=gin.REQUIRED,
               optimizer_fn=None):
    """Constructs a TwoStageModel.

    Args:
      embedding_model_class: Either `values`, `onehot`, or a class that has a
        __call__ function that takes as input a two-tuple of
        (batch_size, num_nodes, heigh, width, num_channels) tensors and returns
        two (batch_size, num_nodes, num_embedding_dims) tensors for both the
        context panels and the answer panels.
      reasoning_model_class: Class that has a __call__ function that takes as
        input a two-tuple of (batch_size, num_nodes, num_embedding_dims) tensors
        and returns the solution in a (batch_size,) tensor.
      optimizer_fn: Function that creates a tf.train optimizer.
    """
    if optimizer_fn is None:
      optimizer_fn = tf.train.AdamOptimizer
    self.optimizer_fn = optimizer_fn
    self.embedding_model_class = embedding_model_class
    self.reasoning_model_class = reasoning_model_class

  def model_fn(self, features, labels, mode, params):
    """TPUEstimator compatible model_fn."""
    del params
    is_training = (mode == tf.estimator.ModeKeys.TRAIN)
    update_ops = []

    # First, embed the context and answer panels.
    if self.embedding_model_class == "values":
      # Use the integer values of the ground-truth factors.
      context_embeddings = features["context_factor_values"]
      answer_embeddings = features["answers_factor_values"]
    elif self.embedding_model_class == "onehot":
      # Use one-hot embeddings of the ground-truth factors.
      context_embeddings = features["context_factors_onehot"]
      answer_embeddings = features["answers_factors_onehot"]
    else:
      embedding_model = self.embedding_model_class()
      context_embeddings, answer_embeddings = embedding_model(
          [
              features["context"],
              features["answers"],
          ],
          training=is_training,
      )
      embedding_model.summary(print_fn=tf.logging.info)
      update_ops += embedding_model.updates

    # Apply the reasoning model.
    reasoning_model = self.reasoning_model_class()
    logits = reasoning_model([context_embeddings, answer_embeddings],
                             training=is_training)
    reasoning_model.summary(print_fn=tf.logging.info)
    update_ops += reasoning_model.updates

    loss_vec = tf.nn.sparse_softmax_cross_entropy_with_logits(
        labels=labels, logits=logits)
    loss_mean = tf.reduce_mean(loss_vec)

    if mode == tf.estimator.ModeKeys.EVAL:

      def metric_fn(labels, logits):
        predictions = tf.argmax(logits, 1)
        return {
            "accuracy":
                tf.metrics.accuracy(labels=labels, predictions=predictions),
        }

      return contrib_tpu.TPUEstimatorSpec(
          mode=mode, loss=loss_mean, eval_metrics=(metric_fn, [labels, logits]))

    if mode == tf.estimator.ModeKeys.TRAIN:
      # In case we use batch norm, the following is required.
      with tf.control_dependencies(update_ops):
        optimizer = self.optimizer_fn()
        train_op = optimizer.minimize(
            loss=loss_mean, global_step=tf.train.get_global_step())
      return contrib_tpu.TPUEstimatorSpec(
          mode=mode, loss=loss_mean, train_op=train_op)
    raise NotImplementedError("Unsupported mode.")


@gin.configurable
class BaselineCNNEmbedder(tf.keras.Model):
  """Baseline implementation where a CNN is learned from scratch."""

  def __init__(self,
               num_latent=gin.REQUIRED,
               name="BaselineCNNEmbedder",
               **kwargs):
    """Constructs a BaselineCNNEmbedder.

    Args:
      num_latent: Integer with the number of latent dimensions.
      name: String with the name of the model.
      **kwargs: Other keyword arguments passed to tf.keras.Model.
    """
    super(BaselineCNNEmbedder, self).__init__(name=name, **kwargs)
    embedding_layers = [
        tf.keras.layers.Conv2D(
            32, (4, 4),
            2,
            activation=get_activation(),
            padding="same",
            kernel_initializer=get_kernel_initializer()),
        tf.keras.layers.Conv2D(
            32, (4, 4),
            2,
            activation=get_activation(),
            padding="same",
            kernel_initializer=get_kernel_initializer()),
        tf.keras.layers.Conv2D(
            64, (4, 4),
            2,
            activation=get_activation(),
            padding="same",
            kernel_initializer=get_kernel_initializer()),
        tf.keras.layers.Conv2D(
            64, (4, 4),
            2,
            activation=get_activation(),
            padding="same",
            kernel_initializer=get_kernel_initializer()),
        tf.keras.layers.Flatten(),
    ]
    self.embedding_layer = relational_layers.MultiDimBatchApply(
        tf.keras.models.Sequential(embedding_layers, "embedding_cnn"))

  def call(self, inputs, **kwargs):
    context, answers = inputs
    context_embedding = self.embedding_layer(context, **kwargs)
    answers_embedding = self.embedding_layer(answers, **kwargs)
    return context_embedding, answers_embedding


@gin.configurable
class HubEmbedding(tf.keras.Model):
  """Embed images using a pre-trained TFHub model.

  Compatible with the representation of a disentanglement_lib model.
  """

  def __init__(self, hub_path=gin.REQUIRED, name="HubEmbedding", **kwargs):
    """Constructs a HubEmbedding.

    Args:
      hub_path: Path to the TFHub module.
      name: String with the name of the model.
      **kwargs: Other keyword arguments passed to tf.keras.Model.
    """
    super(HubEmbedding, self).__init__(name=name, **kwargs)

    def _embedder(x):
      embedder_module = hub.Module(hub_path)
      return embedder_module(dict(images=x), signature="representation")

    self.embedding_layer = relational_layers.MultiDimBatchApply(
        tf.keras.layers.Lambda(_embedder))

  def call(self, inputs, **kwargs):
    context, answers = inputs
    context_embedding = self.embedding_layer(context, **kwargs)
    answers_embedding = self.embedding_layer(answers, **kwargs)
    return context_embedding, answers_embedding


@gin.configurable
class OptimizedWildRelNet(tf.keras.Model):
  """Optimized implementation of the reasoning module in the WildRelNet model.

  Based on https://arxiv.org/pdf/1807.04225.pdf.
  """

  def __init__(self,
               edge_mlp=gin.REQUIRED,
               graph_mlp=gin.REQUIRED,
               dropout_in_last_graph_layer=gin.REQUIRED,
               name="OptimizedWildRelNet",
               **kwargs):
    """Constructs a OptimizedWildRelNet.

    Args:
      edge_mlp: List with number of latent nodes in different layers of the edge
        MLP.
      graph_mlp: List with number of latent nodes in different layers of the
        graph MLP.
      dropout_in_last_graph_layer: Dropout fraction to be applied in the last
        layer of the graph MLP.
      name: String with the name of the model.
      **kwargs: Other keyword arguments passed to tf.keras.Model.
    """
    super(OptimizedWildRelNet, self).__init__(name=name, **kwargs)

    # Create the EdgeMLP.
    edge_layers = []
    for num_units in edge_mlp:
      edge_layers += [
          tf.keras.layers.Dense(
              num_units,
              activation=get_activation(),
              kernel_initializer=get_kernel_initializer())
      ]
    self.edge_layer = tf.keras.models.Sequential(edge_layers, "edge_mlp")

    # Create the GraphMLP.
    graph_layers = []
    for num_units in graph_mlp:
      graph_layers += [
          tf.keras.layers.Dense(
              num_units,
              activation=get_activation(),
              kernel_initializer=get_kernel_initializer())
      ]
    if dropout_in_last_graph_layer:
      graph_layers += [
          tf.keras.layers.Dropout(
              1. - dropout_in_last_graph_layer,
              noise_shape=[1, 1, graph_mlp[-1]])
      ]
    graph_layers += [
        tf.keras.layers.Dense(1, kernel_initializer=get_kernel_initializer())
    ]

    # Create the auxiliary layers.
    self.graph_layer = tf.keras.models.Sequential(graph_layers, "graph_mlp")
    self.stacking_layer = relational_layers.StackAnswers()

    # Create the WildRelationNet.
    self.wildrelnet = tf.keras.models.Sequential([
        relational_layers.AddPositionalEncoding(),
        relational_layers.RelationalLayer(
            self.edge_layer,
            tf.keras.layers.Lambda(lambda x: tf.reduce_sum(x, axis=-2))),
        tf.keras.layers.Lambda(lambda x: tf.reduce_sum(x, axis=-2)),
        self.graph_layer,
        tf.keras.layers.Lambda(lambda x: tf.reduce_sum(x, axis=-1)),
    ], "wildrelnet")

  def call(self, inputs, **kwargs):
    context_embeddings, answer_embeddings = inputs
    # The stacking layer `stacks` each answer panel embedding onto the context
    # panels separately.
    stacked_answers = self.stacking_layer(
        [context_embeddings, answer_embeddings])
    # Apply the relational neural network.
    return self.wildrelnet(stacked_answers, **kwargs)


@gin.configurable("activation")
def get_activation(activation=tf.keras.activations.relu):
  if activation == "lrelu":
    return lambda x: tf.keras.activations.relu(x, alpha=0.2)
  return activation


@gin.configurable("kernel_initializer")
def get_kernel_initializer(kernel_initializer="lecun_normal"):
  return kernel_initializer



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Library of Keras layers used to build relational neural networks."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow.compat.v1 as tf


class RelationalLayer(tf.keras.layers.Layer):
  """Implements a single relational layer.

  This layer is based on equation (1) of https://arxiv.org/pdf/1706.01427.pdf
  and applies (in order) three layers to the input: PairwiseEdgeEmbeddings,
  the specified edge_layer and the reduce_layer. The input tensor and output
  tensor of this layer are of shape (batch_size, num_nodes, num_dims).
  """

  def __init__(self, edge_layer, reduce_layer):
    """Constructs a RelationalLayer.

    Args:
      edge_layer: tf.keras.layers.Layer that is applied to the edge embeddings.
        Should accept a (batch_size, num_nodes, num_nodes, 2*num_dims)-sized
        tensor (where the node embeddings are stacked along the last axis)
        and should return a (batch_size, num_nodes, num_nodes, num_dims)-sized
        tensor.
      reduce_layer: tf.keras.layers.Layer that is used to reduce the edge
        embeddings back to node embeddings. Should accept a
        (batch_size, num_nodes, num_nodes, num_dims)-sized tensor and return a
        (batch_size, num_nodes, num_dims)-sized tensor.
    """
    self._pairwise_edge_embeddings_layer = PairwiseEdgeEmbeddings()
    self.edge_layer = edge_layer
    self.reduce_layer = reduce_layer
    super(RelationalLayer, self).__init__()

  def call(self, inputs, **kwargs):
    # Create edge embedding by concatenating the node embeddings.
    x = self._pairwise_edge_embeddings_layer(inputs, **kwargs)
    # Update edge embeddings using edge layer.
    x = self.edge_layer(x, **kwargs)
    # Reduce back to node embeddings.
    outputs = self.reduce_layer(x, **kwargs)
    return outputs


class PairwiseEdgeEmbeddings(tf.keras.layers.Layer):
  """Creates pairwise edge embeddings from node embeddings."""

  def call(self, inputs, **kwargs):
    num_nodes = inputs.get_shape().as_list()[-2]
    return tf.concat(
        (repeat(tf.expand_dims(inputs, -2), num_nodes, -2),
         repeat(tf.expand_dims(inputs, -3), num_nodes, -3)),
        axis=-1)


def repeat(tensor, num, axis):
  """Repeats tensor num times along the specified axis."""
  multiples = [1] * tensor.get_shape().ndims
  multiples[axis] = num
  return tf.tile(tensor, multiples)


def positional_encoding_like(tensor, positional_encoding_axis=-2,
                             value_axis=-1):
  """Creates positional encoding matching the provided tensor.

  Let each slice along the last axis of the tensor be a row. This function
  computes the index of each row with respect to the specified
  positional_encoding_axis and returns this index using a one-hot embedding.

  The resulting tensor has the same shape as the provided tensor except for the
  value_axis dimension. That dimension contains a one-hot encoding of the
  positional_encoding_axis, i.e., each slice along value_axis and the
  positional_encoding_axis corresponds to the identity matrix.

  Args:
    tensor: Input tensor.
    positional_encoding_axis: Integer with the axis to encode.
    value_axis: Integer with axis where to one-hot encode the
      positional_encoding_axis.

  Returns:
    Positional encoding tensor of the same dtype as tensor.
  """
  # First, create identity matrix for one-hot embedding.
  shape = tensor.get_shape()
  num_values = shape.as_list()[positional_encoding_axis]
  result = tf.eye(num_values, dtype=tensor.dtype)
  # Second, reshape to proper dimensionality.
  new_shape = [1] * shape.ndims
  new_shape[positional_encoding_axis] = num_values
  new_shape[value_axis] = num_values
  result = tf.reshape(result, new_shape)
  # Third, broadcast to final shape.
  multiplier = shape.as_list()
  multiplier[positional_encoding_axis] = 1
  multiplier[value_axis] = 1
  if not shape.is_fully_defined():
    dynamic_shape = tf.unstack(tf.shape(tensor))
    multiplier = [
        (dynamic_shape[i] if n is None else n) for i, n in enumerate(multiplier)
    ]
  return tf.tile(result, multiplier)


class AddPositionalEncoding(tf.keras.layers.Layer):
  """Adds positional encoding to embedding based on position along axis.

  Consider a tensor of shape (batch_size, num_pos, num_dim). This layer will
  return a tensor of shape (batch_size, num_pos, num_dim + num_pos) where
  the slice [:, :, :num_dim] contains the original data while each slice
  [i, :, num_dim:] is the identity matrix. The idea is that this adds extra
  dimensions along axis -1 that encode the position of the element along
  axis -2.
  """

  def __init__(self, positional_encoding_axis=-2, embedding_axis=-1):
    self.positional_encoding_axis = positional_encoding_axis
    self.embedding_axis = embedding_axis
    super(AddPositionalEncoding, self).__init__()

  def call(self, inputs, **kwargs):
    onehot_embedding = positional_encoding_like(
        inputs,
        positional_encoding_axis=self.positional_encoding_axis,
        value_axis=self.embedding_axis)
    return tf.concat([inputs, onehot_embedding], axis=self.embedding_axis)


class StackAnswers(tf.keras.layers.Layer):
  """Concatenates each each answer with the context and stacks them."""

  def __init__(self, answer_axis=-2, stack_axis=-3):
    super(StackAnswers, self).__init__()
    self.answer_axis = answer_axis
    self.stack_axis = stack_axis

  def call(self, inputs, **kwargs):
    context, answers = inputs
    num_answers = answers.get_shape().as_list()[self.answer_axis]
    answer_blocks = []
    for i in range(num_answers):
      ith_answer = tf.gather(answers, [i], axis=self.answer_axis)
      ith_answer_block = tf.concat([context, ith_answer], axis=self.answer_axis)
      answer_blocks.append(ith_answer_block)
    return tf.stack(answer_blocks, axis=self.stack_axis)


class MultiDimBatchApply(tf.keras.layers.Layer):
  """Applies layer for each element in a multi-dimensional batch."""

  def __init__(self, layer, num_dims_to_keep=3):
    """Constructs a MultiDimBatchApply.

    Args:
      layer: tf.keras.layers.Layer to apply.
      num_dims_to_keep: Integer with number of dimensions to provide to the
        layer. The dimensions 0:-num_dims_to_keep correspond to the multi
          dimensional mini batch, i.e., the layer is applied independently for
          each of these elements.
    """
    self.layer = layer
    self.num_dims_to_keep = num_dims_to_keep
    super(MultiDimBatchApply, self).__init__()

  def call(self, inputs, **kwargs):
    shape = inputs.get_shape().as_list()
    collapsed_shape = [-1] + shape[-self.num_dims_to_keep:]
    inputs = tf.reshape(inputs, collapsed_shape)
    # Apply the layer.
    output = self.layer(inputs, **kwargs)
    active_shape = output.get_shape().as_list()[1:]
    output_shape = [-1] + shape[1:-self.num_dims_to_keep] + active_shape
    return tf.reshape(output, output_shape)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main training protocol used for unsupervised disentanglement models."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
import time
from disentanglement_lib.data.ground_truth import named_data
from disentanglement_lib.data.ground_truth import util
from disentanglement_lib.methods.unsupervised import gaussian_encoder_model
from disentanglement_lib.methods.unsupervised import vae  # pylint: disable=unused-import
from disentanglement_lib.utils import results
import numpy as np
import tensorflow.compat.v1 as tf
import gin.tf.external_configurables  # pylint: disable=unused-import
import gin.tf
from tensorflow.contrib import tpu as contrib_tpu


def train_with_gin(model_dir,
                   overwrite=False,
                   gin_config_files=None,
                   gin_bindings=None):
  """Trains a model based on the provided gin configuration.

  This function will set the provided gin bindings, call the train() function
  and clear the gin config. Please see train() for required gin bindings.

  Args:
    model_dir: String with path to directory where model output should be saved.
    overwrite: Boolean indicating whether to overwrite output directory.
    gin_config_files: List of gin config files to load.
    gin_bindings: List of gin bindings to use.
  """
  if gin_config_files is None:
    gin_config_files = []
  if gin_bindings is None:
    gin_bindings = []
  gin.parse_config_files_and_bindings(gin_config_files, gin_bindings)
  train(model_dir, overwrite)
  gin.clear_config()


@gin.configurable("model", blacklist=["model_dir", "overwrite"])
def train(model_dir,
          overwrite=False,
          model=gin.REQUIRED,
          training_steps=gin.REQUIRED,
          random_seed=gin.REQUIRED,
          batch_size=gin.REQUIRED,
          eval_steps=1000,
          name="",
          model_num=None):
  """Trains the estimator and exports the snapshot and the gin config.

  The use of this function requires the gin binding 'dataset.name' to be
  specified as that determines the data set used for training.

  Args:
    model_dir: String with path to directory where model output should be saved.
    overwrite: Boolean indicating whether to overwrite output directory.
    model: GaussianEncoderModel that should be trained and exported.
    training_steps: Integer with number of training steps.
    random_seed: Integer with random seed used for training.
    batch_size: Integer with the batch size.
    eval_steps: Optional integer with number of steps used for evaluation.
    name: Optional string with name of the model (can be used to name models).
    model_num: Optional integer with model number (can be used to identify
      models).
  """
  # We do not use the variables 'name' and 'model_num'. Instead, they can be
  # used to name results as they will be part of the saved gin config.
  del name, model_num

  # Delete the output directory if it already exists.
  if tf.gfile.IsDirectory(model_dir):
    if overwrite:
      tf.gfile.DeleteRecursively(model_dir)
    else:
      raise ValueError("Directory already exists and overwrite is False.")

  # Create a numpy random state. We will sample the random seeds for training
  # and evaluation from this.
  random_state = np.random.RandomState(random_seed)

  # Obtain the dataset.
  dataset = named_data.get_named_ground_truth_data()

  # We create a TPUEstimator based on the provided model. This is primarily so
  # that we could switch to TPU training in the future. For now, we train
  # locally on GPUs.
  run_config = contrib_tpu.RunConfig(
      tf_random_seed=random_seed,
      keep_checkpoint_max=1,
      tpu_config=contrib_tpu.TPUConfig(iterations_per_loop=500))
  tpu_estimator = contrib_tpu.TPUEstimator(
      use_tpu=False,
      model_fn=model.model_fn,
      model_dir=os.path.join(model_dir, "tf_checkpoint"),
      train_batch_size=batch_size,
      eval_batch_size=batch_size,
      config=run_config)

  # Set up time to keep track of elapsed time in results.
  experiment_timer = time.time()

  # Do the actual training.
  tpu_estimator.train(
      input_fn=_make_input_fn(dataset, random_state.randint(2**32)),
      steps=training_steps)

  # Save model as a TFHub module.
  output_shape = named_data.get_named_ground_truth_data().observation_shape
  module_export_path = os.path.join(model_dir, "tfhub")
  gaussian_encoder_model.export_as_tf_hub(model, output_shape,
                                          tpu_estimator.latest_checkpoint(),
                                          module_export_path)

  # Save the results. The result dir will contain all the results and config
  # files that we copied along, as we progress in the pipeline. The idea is that
  # these files will be available for analysis at the end.
  results_dict = tpu_estimator.evaluate(
      input_fn=_make_input_fn(
          dataset, random_state.randint(2**32), num_batches=eval_steps))
  results_dir = os.path.join(model_dir, "results")
  results_dict["elapsed_time"] = time.time() - experiment_timer
  results.update_result_directory(results_dir, "train", results_dict)


def _make_input_fn(ground_truth_data, seed, num_batches=None):
  """Creates an input function for the experiments."""

  def load_dataset(params):
    """TPUEstimator compatible input fuction."""
    dataset = util.tf_data_set_from_ground_truth_data(ground_truth_data, seed)
    batch_size = params["batch_size"]
    # We need to drop the remainder as otherwise we lose the batch size in the
    # tensor shape. This has no effect as our data set is infinite.
    dataset = dataset.batch(batch_size, drop_remainder=True)
    if num_batches is not None:
      dataset = dataset.take(num_batches)
    return dataset.make_one_shot_iterator().get_next()

  return load_dataset



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to make dendrogram plots based on the importance matrices.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import matplotlib
matplotlib.use("Agg")  # Set headless-friendly backend.
import matplotlib.pyplot as plt  # pylint: disable=g-import-not-at-top
import numpy as np
import pandas as pd
from scipy.cluster import hierarchy
import tensorflow.compat.v1 as tf


def dendrogram_plot(matrix, output_dir, factor_names):
  """Make dendrogram plot recording at which threshold factors and codes merge.

  This plotting function produce a dendrogram plot recording at which factors of
  variation and latent codes are most related by running the union-find
  algorithm https://en.wikipedia.org/wiki/Disjoint-set_data_structure on the
  matrix relating factors of variation and latent codes.

  Args:
    matrix: Input matrix of shape [num_factors, num_codes] encoding the
      statistical relation between factors and codes.
    output_dir: Directory to save the plot in.
    factor_names: Lables for the factors of variation to be used in the plot.

  Returns:
    Dictionary containing the threshold ID of each merging events and which
    factors were merged.
  """
  tmp = pd.melt(pd.DataFrame(matrix).reset_index(), id_vars="index")
  # The columns of the dataframe are: index, variable and value.
  tmp = tmp.to_numpy()
  # Sort the matrix by threshold
  tmp = tmp[tmp[:, -1].argsort()[::-1]]
  # The codes have index code + num_factors.
  tmp[:, 1] += matrix.shape[0]

  # Initialize dictionaries for cluster IDs and size.
  size = {}
  cluster_id = {}
  for i in range(matrix.shape[0]):
    size[i] = 1
    cluster_id[i] = i
  # Initialize dendrogram matrix. Each row is an event, each event is composed
  # by [cluster_id_1, cluster_id_2, threshold, size of the new cluster]
  z = np.zeros([matrix.shape[0]-1, 4])
  # Each factor of variation is in its own tree. So the maximum cluster ID we
  # have is matrix.shape[0]-1.
  n_clusters = matrix.shape[0]-1
  nodes = list(range(matrix.shape[0] + matrix.shape[1]))
  idx_found = 0
  discovered = {}
  # Run the Union-Find Algorithm
  for id_i, i in enumerate(tmp):
    # Record if we just discovered a new factor of variation.
    if i[0] not in discovered:
      discovered[i[0]] = id_i
    # Merge trees.
    z, cluster_id, size, n_clusters, idx_found = _union(
        nodes, i[0], i[1], id_i, z, cluster_id, size, matrix, n_clusters,
        idx_found)
  # Obtain the dendrogram plot data structure from the matrix z
  fig, ax = plt.subplots()
  dn = hierarchy.dendrogram(z, ax=ax, orientation="left", no_plot=True)
  # Create a dictionary to map the location on the plot to the leaf
  id_to_leaf = {}
  id_conv = 5
  for l in dn["leaves"]:
    id_to_leaf[id_conv] = l
    id_conv += 10
  # Update the dcoord to when the cluster was actually discovered.
  for d, i in zip(dn["dcoord"], dn["icoord"]):
    if d[0] == 0:
      idx = id_to_leaf[i[0]]
      d[0] = discovered[idx]
    if d[-1] == 0:
      idx = id_to_leaf[i[-1]]
      d[-1] = discovered[idx]
  # Set colors to be all the same.
  dn["color_list"] = ["b"]*len(dn["color_list"])
  dn["ivl"] = np.array(factor_names)[dn["leaves"]]

  hierarchy._plot_dendrogram(dn["icoord"], dn["dcoord"], dn["ivl"], p=30,  # pylint: disable=protected-access
                             n=z.shape[0] + 1, mh=max(z[:, 2]),
                             orientation="right", no_labels=False,
                             color_list=dn["color_list"],
                             leaf_font_size=None,
                             leaf_rotation=None,
                             contraction_marks=None,
                             ax=ax,
                             above_threshold_color="b")
  plt.xlabel("Threshold")
  plt.ylabel("Factor")
  thresholds = tmp[:, 2]
  thresholds_ids = range(0, thresholds.shape[0], 10)
  plt.xticks(
      thresholds_ids,
      np.around(np.array(thresholds, dtype="float32")[thresholds_ids], 2))
  output_path = output_dir+".png"
  with tf.gfile.Open(output_path, "wb") as path:
    fig.savefig(path, bbox_inches="tight")
  return report_merges(z, matrix.shape[0])


def report_merges(z, num_factors):
  """Saves which factors of variations are merged and at which threshold.

  Args:
    z: Dendrogram matrix. Each row is an event, each event is composed by
      [cluster_id_1, cluster_id_2, threshold, size of the new cluster].
    num_factors: Number of factors of Variations.

  Returns:
    Dictionary containing the threshold ID of each merging events and which
    factors were merged.
  """
  scores = {}
  id_to_node_list = {}
  n_clusters = num_factors-1
  for i in range(num_factors):
    id_to_node_list[i] = [i,]
  for i in range(z.shape[0]):
    cluster_id_1 = z[i, 0]
    cluster_id_2 = z[i, 1]
    threshold_id = z[i, 2]
    list_nodes_1 = id_to_node_list[cluster_id_1]
    list_nodes_2 = id_to_node_list[cluster_id_2]

    for node_1 in list_nodes_1:
      for node_2 in list_nodes_2:
        scores["merge_{}_{}".format(node_1, node_2)] = threshold_id

    del id_to_node_list[cluster_id_1]
    del id_to_node_list[cluster_id_2]
    n_clusters += 1

    id_to_node_list[n_clusters] = list_nodes_1 + list_nodes_2

  return scores


def _find(nodes, i):
  """Find function for the Union-Find algorithm."""
  if nodes[i] != i:
    nodes[i] = _find(nodes, nodes[i])
  return nodes[i]


def _union(nodes, idx, idy, val, z, cluster_id, size, matrix, n_clusters,
           idx_found):
  """Implements the a modification to the union of the Union-Find algorithm.

  In this function we first perform the standard union of the Union-Find
  algorithm. We mantain the root of the trees with more than 1 element to
  factors of variation. If two trees rooted at factors of variation gets merged
  we record the event in the dendrogram matrix.

  Args:
    nodes: Array with the nodes of the graph. The first num_factors nodes
      correspond to factors of variation. The rest to codes.
    idx: First node to eventually merge.
    idy: Second node to eventually merge.
    val: Threshold value of the considered edge.
    z: Dendrogram matrix. Each row is an event, each event is composed by
      [cluster_id_1, cluster_id_2, threshold, size of the new cluster].
    cluster_id: Dictionary mapping factors of variation to cluster IDs.
    size: Dictionary mapping cluster ID to its size.
    matrix: Matrix of shape  [num_factors, num_codes] on which we compute the
      dendrogram.
    n_clusters: How many clusters have been discovered.
    idx_found: How many evenys have been found.

  Returns:
    z, cluster_id, size, n_clusters, idx_found
  """
  parent_idx = _find(nodes, idx)
  parent_idy = _find(nodes, idy)
  # Set the parent node the one with smalles id. This ensures that the trees
  # roots are on the factors of variations as they have smaller ID. The two
  # nodes cause a merge if they are not already in the same tree.
  if parent_idx != parent_idy:
    if parent_idy > parent_idx:
      nodes[parent_idy] = parent_idx
    else:
      nodes[parent_idx] = parent_idy

    if parent_idx < matrix.shape[0] and parent_idy < matrix.shape[0]:
      # There is a merge happening on two trees rooted at a factor of variation.
      # These are the events we need to record. First, we get the cluster id of
      # the trees:
      cc_idx = cluster_id[parent_idx]
      cc_idy = cluster_id[parent_idy]
      # Now we create a new cluster by merging the two trees. The size will be
      # the sum of the two subtrees as they are disjoint.
      n_clusters += 1
      size[n_clusters] = size[cc_idx] + size[cc_idy]
      # Update the dendrogram matrix with the new event.
      z[idx_found, :] = [cc_idx, cc_idy, val, size[n_clusters]]
      idx_found += 1
      # Set the new cluster ID for the parent nodes.
      cluster_id[parent_idy] = n_clusters
      cluster_id[parent_idx] = n_clusters
  return z, cluster_id, size, n_clusters, idx_found



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for train.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import absltest
from absl.testing import parameterized
from disentanglement_lib.methods.unsupervised import train
from disentanglement_lib.utils import resources
import gin.tf


def _config_generator():
  """Yields all model configurations that should be tested."""
  model_config_path = resources.get_file(
      "config/tests/methods/unsupervised/train_test.gin")
  # Test different losses.
  for loss in ["@bernoulli_loss", "@l2_loss"]:
    rec_loss = ["reconstruction_loss.loss_fn = " + loss]
    # Test different activations.
    for act in ["'logits'", "'tanh'"]:
      rec_loss += ["reconstruction_loss.activation = " + act]
      latent_dim = ["encoder.num_latent = 10"]
      # Test different architectures.
      for encoder, decoder in [("@fc_encoder", "@fc_decoder"),
                               ("@conv_encoder", "@deconv_decoder")]:
        architectures = \
            ["encoder.encoder_fn = " + encoder,
             "decoder.decoder_fn = " + decoder]
        structure = rec_loss + architectures + latent_dim
        # Train a BetaVAE with all these settings.
        beta_vae = ["model.model = @vae()", "vae.beta = 10."]
        yield [model_config_path], beta_vae + structure

  # Test all the other different models.
  # Test AnnealedVAE.
  annealed_vae = [
      "model.model = @annealed_vae()", "annealed_vae.c_max = 25",
      "annealed_vae.iteration_threshold = 100000", "annealed_vae.gamma = 1000"
  ]
  yield [model_config_path], annealed_vae

  # Test FactorVAE.
  factor_vae = [
      "model.model = @factor_vae()",
      "discriminator.discriminator_fn = @fc_discriminator",
      "discriminator_optimizer.optimizer_fn = @AdamOptimizer",
      "factor_vae.gamma = 10."
  ]
  yield [model_config_path], factor_vae

  # Test DIP-VAE.
  dip_vae_i = [
      "model.model = @dip_vae()", "dip_vae.lambda_d_factor = 10",
      "dip_vae.dip_type = 'i'", "dip_vae.lambda_od = 10."
  ]
  yield [model_config_path], dip_vae_i

  dip_vae_ii = [
      "model.model = @dip_vae()", "dip_vae.lambda_d_factor = 1",
      "dip_vae.dip_type = 'ii'", "dip_vae.lambda_od = 10."
  ]
  yield [model_config_path], dip_vae_ii

  # Test BetaTCVAE.
  beta_tc_vae = ["model.model = @beta_tc_vae()", "beta_tc_vae.beta = 10."]
  yield [model_config_path], beta_tc_vae


class TrainTest(parameterized.TestCase):

  @parameterized.parameters(list(_config_generator()))
  def test_train_model(self, gin_configs, gin_bindings):
    # We clear the gin config before running. Otherwise, if a prior test fails,
    # the gin config is locked and the current test fails.
    gin.clear_config()
    train.train_with_gin(self.create_tempdir().full_path, True, gin_configs,
                         gin_bindings)


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DSprites dataset and new variants with probabilistic decoders."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
from disentanglement_lib.data.ground_truth import ground_truth_data
from disentanglement_lib.data.ground_truth import util
import numpy as np
import PIL
from six.moves import range
from tensorflow.compat.v1 import gfile

DSPRITES_PATH = os.path.join(
    os.environ.get("DISENTANGLEMENT_LIB_DATA", "."), "dsprites",
    "dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz")
SCREAM_PATH = os.path.join(
    os.environ.get("DISENTANGLEMENT_LIB_DATA", "."), "scream", "scream.jpg")



class DSprites(ground_truth_data.GroundTruthData):
  """DSprites dataset.

  The data set was originally introduced in "beta-VAE: Learning Basic Visual
  Concepts with a Constrained Variational Framework" and can be downloaded from
  https://github.com/deepmind/dsprites-dataset.

  The ground-truth factors of variation are (in the default setting):
  0 - shape (3 different values)
  1 - scale (6 different values)
  2 - orientation (40 different values)
  3 - position x (32 different values)
  4 - position y (32 different values)
  """

  def __init__(self, latent_factor_indices=None):
    # By default, all factors (including shape) are considered ground truth
    # factors.
    if latent_factor_indices is None:
      latent_factor_indices = list(range(6))
    self.latent_factor_indices = latent_factor_indices
    self.data_shape = [64, 64, 1]
    # Load the data so that we can sample from it.
    with gfile.Open(DSPRITES_PATH, "rb") as data_file:
      # Data was saved originally using python2, so we need to set the encoding.
      data = np.load(data_file, encoding="latin1", allow_pickle=True)
      self.images = np.array(data["imgs"])
      self.factor_sizes = np.array(
          data["metadata"][()]["latents_sizes"], dtype=np.int64)
    self.full_factor_sizes = [1, 3, 6, 40, 32, 32]
    self.factor_bases = np.prod(self.factor_sizes) / np.cumprod(
        self.factor_sizes)
    self.state_space = util.SplitDiscreteStateSpace(self.factor_sizes,
                                                    self.latent_factor_indices)

  @property
  def num_factors(self):
    return self.state_space.num_latent_factors

  @property
  def factors_num_values(self):
    return [self.full_factor_sizes[i] for i in self.latent_factor_indices]

  @property
  def observation_shape(self):
    return self.data_shape


  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    return self.state_space.sample_latent_factors(num, random_state)

  def sample_observations_from_factors(self, factors, random_state):
    return self.sample_observations_from_factors_no_color(factors, random_state)

  def sample_observations_from_factors_no_color(self, factors, random_state):
    """Sample a batch of observations X given a batch of factors Y."""
    all_factors = self.state_space.sample_all_factors(factors, random_state)
    indices = np.array(np.dot(all_factors, self.factor_bases), dtype=np.int64)
    return np.expand_dims(self.images[indices].astype(np.float32), axis=3)

  def _sample_factor(self, i, num, random_state):
    return random_state.randint(self.factor_sizes[i], size=num)


class ColorDSprites(DSprites):
  """Color DSprites.

  This data set is the same as the original DSprites data set except that when
  sampling the observations X, the sprite is colored in a randomly sampled
  color.

  The ground-truth factors of variation are (in the default setting):
  0 - shape (3 different values)
  1 - scale (6 different values)
  2 - orientation (40 different values)
  3 - position x (32 different values)
  4 - position y (32 different values)
  """

  def __init__(self, latent_factor_indices=None):
    DSprites.__init__(self, latent_factor_indices)
    self.data_shape = [64, 64, 3]

  def sample_observations_from_factors(self, factors, random_state):
    no_color_observations = self.sample_observations_from_factors_no_color(
        factors, random_state)
    observations = np.repeat(no_color_observations, 3, axis=3)
    color = np.repeat(
        np.repeat(
            random_state.uniform(0.5, 1, [observations.shape[0], 1, 1, 3]),
            observations.shape[1],
            axis=1),
        observations.shape[2],
        axis=2)
    return observations * color


class NoisyDSprites(DSprites):
  """Noisy DSprites.

  This data set is the same as the original DSprites data set except that when
  sampling the observations X, the background pixels are replaced with random
  noise.

  The ground-truth factors of variation are (in the default setting):
  0 - shape (3 different values)
  1 - scale (6 different values)
  2 - orientation (40 different values)
  3 - position x (32 different values)
  4 - position y (32 different values)
  """

  def __init__(self, latent_factor_indices=None):
    DSprites.__init__(self, latent_factor_indices)
    self.data_shape = [64, 64, 3]

  def sample_observations_from_factors(self, factors, random_state):
    no_color_observations = self.sample_observations_from_factors_no_color(
        factors, random_state)
    observations = np.repeat(no_color_observations, 3, axis=3)
    color = random_state.uniform(0, 1, [observations.shape[0], 64, 64, 3])
    return np.minimum(observations + color, 1.)


class ScreamDSprites(DSprites):
  """Scream DSprites.

  This data set is the same as the original DSprites data set except that when
  sampling the observations X, a random patch of the Scream image is sampled as
  the background and the sprite is embedded into the image by inverting the
  color of the sampled patch at the pixels of the sprite.

  The ground-truth factors of variation are (in the default setting):
  0 - shape (3 different values)
  1 - scale (6 different values)
  2 - orientation (40 different values)
  3 - position x (32 different values)
  4 - position y (32 different values)
  """

  def __init__(self, latent_factor_indices=None):
    DSprites.__init__(self, latent_factor_indices)
    self.data_shape = [64, 64, 3]
    with gfile.Open(SCREAM_PATH, "rb") as f:
      scream = PIL.Image.open(f)
      scream.thumbnail((350, 274, 3))
      self.scream = np.array(scream) * 1. / 255.

  def sample_observations_from_factors(self, factors, random_state):
    no_color_observations = self.sample_observations_from_factors_no_color(
        factors, random_state)
    observations = np.repeat(no_color_observations, 3, axis=3)

    for i in range(observations.shape[0]):
      x_crop = random_state.randint(0, self.scream.shape[0] - 64)
      y_crop = random_state.randint(0, self.scream.shape[1] - 64)
      background = (self.scream[x_crop:x_crop + 64, y_crop:y_crop + 64] +
                    random_state.uniform(0, 1, size=3)) / 2.
      mask = (observations[i] == 1)
      background[mask] = 1 - background[mask]
      observations[i] = background
    return observations


# Object colors generated using
# >> seaborn.husl_palette(n_colors=6, h=0.1, s=0.7, l=0.7)
OBJECT_COLORS = np.array(
    [[0.9096231780824386, 0.5883403686424795, 0.3657680693481871],
     [0.6350181801577739, 0.6927729880940552, 0.3626904230371999],
     [0.3764832455369271, 0.7283900430001952, 0.5963114605342514],
     [0.39548987063404156, 0.7073922557810771, 0.7874577552076919],
     [0.6963644829189117, 0.6220697032672371, 0.899716387820763],
     [0.90815966835861, 0.5511103319168646, 0.7494337214212151]])

BACKGROUND_COLORS = np.array([
    (0., 0., 0.),
    (.25, .25, .25),
    (.5, .5, .5),
    (.75, .75, .75),
    (1., 1., 1.),
])


class AbstractDSprites(DSprites):
  """DSprites variation for abstract reasoning task.

  Rotation is not considered a ground-truth factor and we sample random colors
  both for the object and the background.

  The ground-truth factors of variation are (in the default setting):
  0 - background color (5 different values)
  1 - object color (6 different values)
  2 - shape (3 different values)
  3 - scale (6 different values)
  4 - position x (32 different values)
  5 - position y (32 different values)
  """

  def __init__(self):
    # We retain all original factors except shape.
    DSprites.__init__(self, [1, 2, 4, 5])
    self.data_shape = [64, 64, 3]

  @property
  def num_factors(self):
    return 2 + self.state_space.num_latent_factors

  @property
  def factors_num_values(self):
    return ([BACKGROUND_COLORS.shape[0], OBJECT_COLORS.shape[0]] +
            [self.full_factor_sizes[i] for i in self.latent_factor_indices])

  def sample_factors(self, num, random_state):
    """Sample a batch of factors Y."""
    colors = np.zeros((num, 2), dtype=np.int64)
    colors[:, 0] = random_state.randint(BACKGROUND_COLORS.shape[0], size=num)
    colors[:, 1] = random_state.randint(OBJECT_COLORS.shape[0], size=num)
    other_factors = self.state_space.sample_latent_factors(num, random_state)
    return np.concatenate([colors, other_factors], axis=-1)

  def sample_observations_from_factors(self, factors, random_state):
    mask = self.sample_observations_from_factors_no_color(
        factors[:, 2:], random_state)

    background_color = BACKGROUND_COLORS[factors[:, 0]]
    object_color = OBJECT_COLORS[factors[:, 1]]

    # Add axis for height and width.
    background_color = np.expand_dims(np.expand_dims(background_color, 1), 1)
    object_color = np.expand_dims(np.expand_dims(object_color, 1), 1)

    return mask * object_color + (1. - mask) * background_color



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the architectures.py."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from absl.testing import parameterized
from disentanglement_lib.methods.shared import architectures
import numpy as np
import tensorflow.compat.v1 as tf


class ArchitecturesTest(parameterized.TestCase, tf.test.TestCase):

  @parameterized.named_parameters(
      ('fc_encoder', architectures.fc_encoder),
      ('conv_encoder', architectures.conv_encoder),
  )
  def test_encoder(self, encoder_f):
    minibatch = np.ones(shape=(10, 64, 64, 1), dtype=np.float32)
    input_tensor = tf.placeholder(tf.float32, shape=(None, 64, 64, 1))
    latent_mean, latent_logvar = encoder_f(input_tensor, 10)
    with self.test_session() as sess:
      sess.run(tf.initialize_all_variables())
      sess.run(
          [latent_mean, latent_logvar], feed_dict={input_tensor: minibatch})

  @parameterized.named_parameters(
      ('fc_decoder', architectures.fc_decoder),
      ('deconv_decoder', architectures.deconv_decoder),
  )
  def test_decoder(self, decoder_f):
    latent_variable = np.ones(shape=(10, 15), dtype=np.float32)
    input_tensor = tf.placeholder(tf.float32, shape=(None, 15))
    images = decoder_f(input_tensor, [64, 64, 1])
    with self.test_session() as sess:
      sess.run(tf.initialize_all_variables())
      sess.run(images, feed_dict={input_tensor: latent_variable})

  @parameterized.named_parameters(
      ('fc_discriminator', architectures.fc_discriminator),
  )
  def test_discriminator(self, discriminator_f):
    images = np.ones(shape=(32, 10), dtype=np.float32)
    input_tensor = tf.placeholder(tf.float32, shape=(None, 10))
    logits, probs = discriminator_f(input_tensor)
    with self.test_session() as sess:
      sess.run(tf.initialize_all_variables())
      sess.run([logits, probs], feed_dict={input_tensor: images})

if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of a unified disentanglement score.

This score generalizes several existing disentanglement scores (DCI, MIG,
Modularity and SAP). First, we compute a matrix relating factors of variations
and latent codes. Then, we aggregate this matrix into a score.
DCI, MIG, Modularity and SAP can be obtained as a special case combining their
matrix with the corresponding aggregation.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from absl import logging
from disentanglement_lib.evaluation.metrics import dci
from disentanglement_lib.evaluation.metrics import modularity_explicitness
from disentanglement_lib.evaluation.metrics import sap_score
from disentanglement_lib.evaluation.metrics import utils
from disentanglement_lib.utils import results
from disentanglement_lib.visualize import dendrogram
from disentanglement_lib.visualize import visualize_scores
import numpy as np

import gin.tf


@gin.configurable(
    "unified_scores",
    blacklist=["ground_truth_data", "representation_function", "random_state",
               "artifact_dir"])
def compute_unified_scores(ground_truth_data, representation_function,
                           random_state,
                           artifact_dir=None,
                           num_train=gin.REQUIRED,
                           num_test=gin.REQUIRED,
                           matrix_fns=gin.REQUIRED,
                           batch_size=16):
  """Computes the unified disentanglement scores.

  Args:
    ground_truth_data: GroundTruthData to be sampled from.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    random_state: Numpy random state used for randomness.
    artifact_dir: Optional path to directory where artifacts can be saved.
    num_train: Number of points used for training.
    num_test: Number of points used for testing.
    matrix_fns: List of functions to relate factors of variations and codes.
    batch_size: Batch size for sampling.

  Returns:
    Unified scores.
  """
  logging.info("Generating training set.")
  # mus_train are of shape [num_codes, num_train], while ys_train are of shape
  # [num_factors, num_train].
  mus_train, ys_train = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_train,
      random_state, batch_size)
  assert mus_train.shape[1] == num_train
  assert ys_train.shape[1] == num_train
  mus_test, ys_test = utils.generate_batch_factor_code(
      ground_truth_data, representation_function, num_test,
      random_state, batch_size)

  return unified_scores(mus_train, ys_train, mus_test, ys_test, matrix_fns,
                        artifact_dir, ground_truth_data.factor_names)


@gin.configurable(
    "unified_score_validation",
    blacklist=["observations", "labels", "representation_function"])
def compute_unified_score_on_fixed_data(
    observations, labels, representation_function,
    train_percentage=gin.REQUIRED, matrix_fns=gin.REQUIRED, batch_size=100):
  """Computes the unified scores on the fixed set of observations and labels.

  Args:
    observations: Observations on which to compute the score. Observations have
      shape (num_observations, 64, 64, num_channels).
    labels: Observed factors of variations.
    representation_function: Function that takes observations as input and
      outputs a dim_representation sized representation for each observation.
    train_percentage: Percentage of observations used for training.
    matrix_fns: List of functions to relate factors of variations and codes.
    batch_size: Batch size used to compute the representation.

  Returns:
    Unified scores.
  """
  mus = utils.obtain_representation(observations, representation_function,
                                    batch_size)
  assert labels.shape[1] == observations.shape[0], "Wrong labels shape."
  assert mus.shape[1] == observations.shape[0], "Wrong representation shape."

  mus_train, mus_test = utils.split_train_test(
      mus,
      train_percentage)
  ys_train, ys_test = utils.split_train_test(
      labels,
      train_percentage)
  return unified_scores(mus_train, ys_train, mus_test, ys_test, matrix_fns)


def unified_scores(mus_train, ys_train, mus_test, ys_test, matrix_fns,
                   artifact_dir=None, factor_names=None):
  """Computes unified scores."""

  scores = {}
  kws = {}
  for matrix_fn in matrix_fns:
    # Matrix should have shape [num_codes, num_factors].
    matrix = matrix_fn(mus_train, ys_train, mus_test, ys_test)
    matrix_name = matrix_fn.__name__
    if artifact_dir is not None:
      visualize_scores.heat_square(matrix.copy(), artifact_dir, matrix_name,
                                   "Latent codes", "Factors of Variation",
                                   factor_names=factor_names)
      visualize_scores.plot_recovery_vs_independent(matrix.copy().T,
                                                    artifact_dir,
                                                    matrix_name+"_pr")
      merge_points = dendrogram.dendrogram_plot(matrix.copy().T, os.path.join(
          artifact_dir, matrix_name+"_dendrogram"), factor_names)
      kws[matrix_name] = merge_points
    results_dict = pr_curves_values(matrix)
    if matrix_name in kws:
      kws[matrix_name].update(results_dict)
    else:
      kws[matrix_name] = results_dict
    for aggregation_fn in AGGREGATION_FN:
      results_dict = aggregation_fn(matrix, ys_train)
      kws[matrix_name].update(results_dict)
  scores = results.namespaced_dict(scores, **kws)
  return scores


def pr_curves_values(matrix):
  """Computes area of precision curve and max recall."""
  scores = {}
  thresholds = np.sort(matrix.flatten())[::-1]
  precisions = [visualize_scores.precision(matrix, x) for x in thresholds]
  area = 0.
  for i in range(len(precisions)-1):
    area += (
        precisions[i] + precisions[i+1])*(thresholds[i] - thresholds[i+1])*0.5
  scores["area_precision"] = area
  scores["max_precision"] = max(precisions)
  return scores


@gin.configurable(
    "importance_gbt_matrix",
    blacklist=["mus_train", "ys_train", "mus_test", "ys_test"])
def importance_gbt_matrix(mus_train, ys_train, mus_test, ys_test):
  """Computes the importance matrix of the DCI Disentanglement score.

  The importance matrix is based on the importance of each code to predict a
  factor of variation with GBT.

  Args:
    mus_train: Batch of learned representations to be used for training.
    ys_train: Observed factors of variation corresponding to the representations
      in mus_train.
    mus_test: Batch of learned representations to be used for testing.
    ys_test: Observed factors of variation corresponding to the representations
    in mus_test.

  Returns:
    Importance matrix as computed for the DCI Disentanglement score.
  """
  matrix_importance_gbt, _, _ = dci.compute_importance_gbt(
      mus_train, ys_train, mus_test, ys_test)
  return matrix_importance_gbt


@gin.configurable(
    "mutual_information_matrix",
    blacklist=["mus_train", "ys_train", "mus_test", "ys_test"])
def mutual_information_matrix(mus_train, ys_train, mus_test, ys_test):
  """Computes the mutual information matrix between codes and factors.

  The mutual information matrix is used to compute the MIG and Modularity
  scores.

  Args:
    mus_train: Batch of learned representations to be used for training.
    ys_train: Observed factors of variation corresponding to the representations
      in mus_train.
    mus_test: Unused.
    ys_test: Unused.

  Returns:
    Mutual information matrix as computed for the MIG and Modularity scores.
  """
  del mus_test, ys_test
  discretized_mus = utils.make_discretizer(mus_train)
  m = utils.discrete_mutual_info(discretized_mus, ys_train)
  return m


@gin.configurable(
    "accuracy_svm_matrix",
    blacklist=["mus_train", "ys_train", "mus_test", "ys_test"])
def accuracy_svm_matrix(mus_train, ys_train, mus_test, ys_test):
  """Prediction accuracy of a SVM predicting a factor from a single code.

  The matrix of accuracies is used to compute the SAP score.

  Args:
    mus_train: Batch of learned representations to be used for training.
    ys_train: Observed factors of variation corresponding to the representations
      in mus_train.
    mus_test: Batch of learned representations to be used for testing.
    ys_test: Observed factors of variation corresponding to the representations
    in mus_test.

  Returns:
    Accuracy matrix as computed for the SAP score.
  """
  return sap_score.compute_score_matrix(
      mus_train, ys_train, mus_test, ys_test, continuous_factors=False)


def aggregation_dci(matrix, ys):
  """Aggregation function of the DCI Disentanglement."""
  del ys
  score = {}
  score["dci_disentanglement"] = dci.disentanglement(matrix)
  score["dci_completeness"] = dci.completeness(matrix)
  score["dci"] = dci.disentanglement(matrix)
  disentanglement_per_code = dci.disentanglement_per_code(matrix)
  completeness_per_factor = dci.completeness_per_factor(matrix)
  assert len(disentanglement_per_code) == matrix.shape[0], "Wrong length."
  assert len(completeness_per_factor) == matrix.shape[1], "Wrong length."
  for i in range(len(disentanglement_per_code)):
    score["dci_disentanglement.code_{}".format(i)] = disentanglement_per_code[i]
  for i in range(len(completeness_per_factor)):
    score["dci_completeness.code_{}".format(i)] = completeness_per_factor[i]
  return score


def aggregation_mig(m, ys_train):
  """Aggregation function of the MIG."""
  score = {}
  entropy = utils.discrete_entropy(ys_train)
  sorted_m = np.sort(m, axis=0)[::-1]
  mig_per_factor = np.divide(sorted_m[0, :] - sorted_m[1, :], entropy[:])
  score["mig"] = np.mean(mig_per_factor)
  assert len(mig_per_factor) == m.shape[1], "Wrong length."
  for i in range(len(mig_per_factor)):
    score["mig.factor_{}".format(i)] = mig_per_factor[i]
  return score


def aggregation_sap(matrix, ys):
  """Aggregation function of the SAP score."""
  del ys
  score = {}
  score["sap"] = sap_score.compute_avg_diff_top_two(matrix)
  sap_per_factor = sap_compute_diff_top_two(matrix)
  assert len(sap_per_factor) == matrix.shape[1], "Wrong length."
  for i in range(len(sap_per_factor)):
    score["sap.factor_{}".format(i)] = sap_per_factor[i]
  return score


def sap_compute_diff_top_two(matrix):
  sorted_matrix = np.sort(matrix, axis=0)
  return sorted_matrix[-1, :] - sorted_matrix[-2, :]


def aggregation_modularity(matrix, ys):
  """Aggregation function of the modularity score."""
  del ys
  score = {}
  score["modularity"] = modularity_explicitness.modularity(matrix)
  modularity_per_code = compute_modularity_per_code(matrix)
  assert len(modularity_per_code) == matrix.shape[0], "Wrong length."
  for i in range(len(modularity_per_code)):
    score["modularity.code_{}".format(i)] = modularity_per_code[i]
  return score


def compute_modularity_per_code(mutual_information):
  """Computes the modularity from mutual information."""
  # Mutual information has shape [num_codes, num_factors].
  squared_mi = np.square(mutual_information)
  max_squared_mi = np.max(squared_mi, axis=1)
  numerator = np.sum(squared_mi, axis=1) - max_squared_mi
  denominator = max_squared_mi * (squared_mi.shape[1] -1.)
  delta = numerator / denominator
  modularity_score = 1. - delta
  index = (max_squared_mi == 0.)
  modularity_score[index] = 0.
  return modularity_score

AGGREGATION_FN = [aggregation_dci, aggregation_mig, aggregation_sap,
                  aggregation_modularity]



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for unified_scores.py."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Dependency imports

from absl.testing import absltest
from disentanglement_lib.data.ground_truth import dummy_data
from disentanglement_lib.evaluation.metrics import strong_downstream_task
from disentanglement_lib.evaluation.metrics import utils
import numpy as np

import gin.tf


def _identity_discretizer(target, num_bins):
  del num_bins
  return target


class StrongDownstreamTaskTest(absltest.TestCase):

  def test_intervene(self):
    ground_truth_data = dummy_data.DummyData()

    random_state = np.random.RandomState(0)

    ys_train = ground_truth_data.sample_factors(1000, random_state)
    ys_test = ground_truth_data.sample_factors(1000, random_state)
    num_factors = ys_train.shape[1]
    for i in range(num_factors):
      (y_train_int, y_test_int, interv_factor,
       factor_interv_train) = strong_downstream_task.intervene(
           ys_train.copy(), ys_test.copy(), i, num_factors, ground_truth_data)
      assert interv_factor != i, "Wrong factor interevened on."
      assert (y_train_int[:, interv_factor] == factor_interv_train
             ).all(), "Training set not intervened on."
      assert (y_test_int[:, interv_factor] != factor_interv_train
             ).all(), "Training set not intervened on."

  def test_task(self):
    gin.bind_parameter("predictor.predictor_fn",
                       utils.gradient_boosting_classifier)
    gin.bind_parameter("strong_downstream_task.num_train",
                       [1000])
    gin.bind_parameter("strong_downstream_task.num_test",
                       1000)
    gin.bind_parameter("strong_downstream_task.n_experiment",
                       2)
    ground_truth_data = dummy_data.DummyData()
    def representation_function(x):
      return np.array(x, dtype=np.float64)[:, :, 0, 0]
    random_state = np.random.RandomState(0)
    scores = strong_downstream_task.compute_strong_downstream_task(
        ground_truth_data, representation_function, random_state,
        artifact_dir=None)
    self.assertBetween(scores["1000:mean_strong_test_accuracy"], 0.0, 0.3)

if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to save results and gin configs in result directory."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import copy
import os
import re
import uuid
from distutils import dir_util
import numpy as np
import simplejson as json
import tensorflow.compat.v1 as tf
from tensorflow.compat.v1 import gfile
import gin.tf


def update_result_directory(result_directory,
                            step_name,
                            results_dict,
                            old_result_directory=None):
  """One stop solution for updating the result directory.

  1. Copies old_result_directory to result_directory if not None.
  2. Adds a unique id to the result_dict.
  3. Saves the gin config to the gin/{step_name}.gin file.
  4. Saves the gin config dict to json/config_{step_name}.json file.
  5. Saves the results_dict to the json/results_{step_name}.json file.
  6. Aggregates all dicts in json/*.json into a new
     aggregate/aggregate_results_{step_name}.json file.

  Args:
    result_directory: String with path to result directory to update.
    step_name: String with the step name. This will be used as a name space.
    results_dict: Dictionary with results to be persisted.
    old_result_directory: String with path to old directory from which to copy
      results from (if not set to None).
  """
  json_dir = os.path.join(result_directory, "json")

  # Copy the old output directory to the new one if required.
  if old_result_directory is not None:
    copydir(old_result_directory, result_directory)
  else:
    # Creates the output directory if necessary.
    if not tf.gfile.IsDirectory(result_directory):
      tf.gfile.MakeDirs(result_directory)

  # Add unique id to the result dict (useful for obtaining unique runs).
  results_dict["uuid"] = str(uuid.uuid4())

  # Save the gin config in the gin format.
  gin_config_path = os.path.join(result_directory, "gin",
                                 "{}.gin".format(step_name))
  save_gin(gin_config_path)

  # Save gin config in JSON file for aggregation.
  gin_json_path = os.path.join(json_dir, "{}_config.json".format(step_name))
  save_dict(gin_json_path, gin_dict(gin_config_path))

  # Save the results as a dictionary in JSON.
  results_json_path = os.path.join(json_dir,
                                   "{}_results.json".format(step_name))
  save_dict(results_json_path, results_dict)

  # Aggregate all the results present in the result_dir so far.
  aggregate_dict = aggregate_json_results(json_dir)
  aggregate_json_path = results_json_path = os.path.join(
      result_directory, "aggregate", "{}.json".format(step_name))
  save_dict(aggregate_json_path, aggregate_dict)


def _copy_recursively(path_to_old_dir, path_to_new_dir):
  return dir_util.copy_tree(path_to_old_dir, path_to_new_dir)


def copydir(path_to_old_dir, path_to_new_dir):
  """Copies a directory to a new path which is created if necessary.

  Args:
    path_to_old_dir: String with old directory path.
    path_to_new_dir: String with new directory path.
  """
  directory = os.path.dirname(path_to_new_dir)
  if not tf.gfile.IsDirectory(directory):
    tf.gfile.MakeDirs(directory)
  _copy_recursively(path_to_old_dir, path_to_new_dir)


def save_gin(config_path):
  """Saves the operative gin config to a gin config file.

  Args:
    config_path: String with path where to save the gin config.
  """
  # Ensure that the folder exists.
  directory = os.path.dirname(config_path)
  if not tf.gfile.IsDirectory(directory):
    tf.gfile.MakeDirs(directory)
  # Save the actual config.
  with tf.gfile.GFile(config_path, "w") as f:
    f.write(gin.operative_config_str())


class Encoder(json.JSONEncoder):
  """Custom encoder so that we can save special types in JSON."""

  def default(self, obj):
    if isinstance(obj, (np.float_, np.float32, np.float16, np.float64)):
      return float(obj)
    elif isinstance(obj,
                    (np.intc, np.intp, np.int_, np.int8, np.int16, np.int32,
                     np.int64, np.uint8, np.uint16, np.uint32, np.uint64)):
      return int(obj)
    elif isinstance(obj, np.ndarray):
      obj = obj.tolist()
    return json.JSONEncoder.default(self, obj)


def save_dict(config_path, dict_with_info):
  """Saves a dict to a JSON file.

  Args:
    config_path: String with path where to save the gin config.
    dict_with_info: Dictionary with keys and values which are safed as strings.
  """
  # Ensure that the folder exists.
  directory = os.path.dirname(config_path)
  if not tf.gfile.IsDirectory(directory):
    tf.gfile.MakeDirs(directory)
  # Save the actual config.
  with tf.gfile.GFile(config_path, "w") as f:
    json.dump(dict_with_info, f, cls=Encoder, indent=2)


def gin_dict(config_path=None):
  """Returns dict with gin configs based on active config or config file.

  Args:
    config_path: Path to gin config file. If set to None (default), currently
      active bindings using gin.operative_config_str() are used.

  Returns:
    Dictionary with gin bindings as string keys and string values.
  """
  result = {}
  # Gin does not allow to directly retrieve a dictionary but it allows to
  # obtain a string with all active configs in human readable format.
  if config_path is None:
    operative_str = gin.operative_config_str()
  else:
    with tf.gfile.GFile(config_path, "r") as f:
      operative_str = f.read()
  for line in operative_str.split("\n"):
    # We need to filter out the auto-generated comments and make sure the line
    # contains a valid assignment.
    if not line.startswith("#") and " = " in line:
      key, value = line.split(" = ", 2)
      result[key] = value
  return result


def namespaced_dict(base_dict=None, **named_dicts):
  """Fuses several named dictionaries into one dict by namespacing the keys.

  Example:
  >> base_dict = {"!": "!!"}
  >> numbers = {"1": "one"}
  >> chars = {"a": "A"}
  >> new_dict = namespaced_dict(base_dict, numbers=numbers, chars=chars)
  >> # new_dict = {"!": "!!", "numbers.1": "one", "chars.a": "A"}

  Args:
    base_dict: Base dictionary of which a deepcopy will be use to fuse the named
      dicts into. If set to None, an empty dict will be used.
    **named_dicts: Named dictionary of dictionaries that will be namespaced and
      fused into base_dict. All keys should be string as the new key of any
      value will be outer key + "." + inner key.

  Returns:
    Dictionary with aggregated items.
  """
  result = {} if base_dict is None else copy.deepcopy(base_dict)
  for outer_key, inner_dict in named_dicts.items():
    for inner_key, value in inner_dict.items():
      result["{}.{}".format(outer_key, inner_key)] = value
  return result


def aggregate_json_results(base_path):
  """Aggregates all the result files in a directory into a namespaced dict.

  Args:
    base_path: String with the directory containing JSON files that only contain
      dictionaries.

  Returns:
    Namespaced dictionary with the results.
  """
  result = {}
  compiled_pattern = re.compile(r"(.*)\.json")
  for filename in gfile.ListDirectory(base_path):
    match = compiled_pattern.match(filename)
    if match:
      path = os.path.join(base_path, filename)
      with tf.gfile.GFile(path, "r") as f:
        result[match.group(1)] = json.load(f)
  return namespaced_dict(**result)



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers to save and load the results of many experiments in a single JSON."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import multiprocessing
from absl import logging
import pandas as pd
import simplejson as json
from tensorflow.compat.v1 import gfile


def aggregate_results_to_json(result_file_pattern, output_path):
  """Aggregates all the results files in the pattern into a single JSON file.

  Args:
    result_file_pattern: String with glob pattern to all the result files that
      should be aggregated (e.g. /tmp/*/results/aggregate/evaluation.json).
    output_path: String with path to output json file (e.g. /tmp/results.json).
  """
  logging.info("Loading the results.")
  model_results = _get(result_file_pattern)
  logging.info("Saving the aggregated results.")
  with gfile.Open(output_path, "w") as f:
    model_results.to_json(path_or_buf=f)


def load_aggregated_json_results(source_path):
  """Convenience function to load aggregated results from JSON file.

  Args:
    source_path: String with path to aggregated json file (e.g.
      /tmp/results.json).

  Returns:
    pd.DataFrame with aggregated results.
  """
  logging.info("Loading the aggregated results.")
  return pd.read_json(path_or_buf=source_path, orient="columns")


def _load(path):
  with gfile.GFile(path) as f:
    result = json.load(f)
  result["path"] = path
  return result


def _get(pattern):
  files = gfile.Glob(pattern)
  pool = multiprocessing.Pool()
  all_results = pool.map(_load, files)
  return pd.DataFrame(all_results)


